# -*- coding: utf-8 -*-
#
# ListBuffer, version 0.8.1 for WeeChat version 0.3
# Latest development version: https://github.com/FiXato/listbuffer
#
#   Show /list results in a common buffer and interact with them.
#
#   This script allows you to easily join channels from the /list output.
#   It will open a common buffer for the /list result, through which you
#   browse with your cursor keys, and join with the meta-enter keys.
#   Adjust sorting with meta->, meta-< and meta-/ keybindings.
#
## History:
### 2011-09-08: FiXato:
#
# * version 0.1:  initial release.
#     * added a common buffer for /list results
#     * added highlighting for currently selected line
#     * added /join support via enter key
#     * added scroll_top and scroll_bottom support
#
# * version 0.2:  /list format bugfix
#     * added support for /list results without modes
#     * some servers don't send 321 (/list start). Taken into account.
#
# * version 0.3: Sorting support
#     * Added some basic sorting support. Scroll through sort options
#       with meta-> and meta-< (users, channel, topic, modes)
#
### 2011-09-19: FiXato
#
# * version 0.4:
#     * Case-insensitive buffer lookup fix.
#     * Removed default enter keybind
#
### 2011-12-28: troydm:
#
# * version 0.5: It's an upside-down-world
#     * Added inverted sorting support provided by Dmitry "troydm" Geurkov
#       Use meta-/ to switch between inverted and regular sorting.
#
### 2012-02-10: FiXato:
#
# * version 0.6: Stop shoving that buffer in my face!
#     * The listbuffer should no longer pop up by itself when you load the script.
#       It should only pop up now when you actually do a /list query.
#
# * version 0.7: .. but please pop it up in my current window when I ask for it
#     * Added setting plugins.var.python.listbuffer.autofocus
#       This will autofocus the listbuffer in the current window if another window isn't
#       already showing it, and of course only when the user issues /list
#
### 2012-07-10: FiXato:
#
# * version 0.7.1: Forgetful bugfix
#     * Made sure lb_curline global variable is defined
#
### 2013-03-19: FiXato:
#
# * version 0.8: Sorted out the sorting
#     * Added automatically updating options for sorting:
#       * plugins.var.python.listbuffer.sort_inverted
#       * plugins.var.python.listbuffer.sort_order
# * version 0.8.1: Pad it baby!
#     * Channel modes are equally padded even when there are no channel modes.
#     * Added padding options:
#       * plugins.var.python.listbuffer.modes_min_width
#       * plugins.var.python.listbuffer.channel_min_width
#       * plugins.var.python.listbuffer.users_min_width
#
### 2019-07-05: Sébastien Helleu:
#
# * version 0.8.2: Make script compatible with Python 3.
#
## Acknowledgements:
# * Dmitry "troydm" Geurkov, for providing the inverse-sorting patch to the project.
# * Sebastien "Flashcode" Helleu, for developing the kick-ass IRC client WeeChat
#    and the iset.pl script which inspired me to this script.
# * Nils "nils_2" Görs, for his contributions to iset.pl which served as
#    example code.
# * David "drubin" Rubin, for his urlgrab.py script, which also served
#    as example code.
# * ArZa, whose listsort.pl script helped me getting started with
#    grabbing the /list results. Parts of his code have been shamelessly
#    copied and ported to Python.
# * Khaled Mardam-Bey, for making me yearn for similar /list support in
#    WeeChat as mIRC already offered. :P
# * mave_, for pointing out that sort orders weren't remembered.
#
## TODO:
#   - Auto-scroll selected line upon window scroll.
#   - Add option to hide already joined channels.
#   - Improve sorting methods
#   - Add auto-join support
#   - Detect if channel is already in auto-join
#   - Allow automatically switching to the listbuffer
#   - Add support for ALIS (/squery alis LIST * -mix 100 (IRCNet)
#   - Make colours configurable
#   - Limit number of channels to parse
#   - Add filter support a la iset
#   - Allow selecting multiple channels
#   - Add optional command redirection.
#
## Copyright (c) 2011,2012,2013 Filip H.F. "FiXato" Slagter,
#   <FiXato [at] Gmail [dot] com>
#   http://profile.fixato.org
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from __future__ import print_function

SCRIPT_NAME    = "listbuffer"
SCRIPT_AUTHOR  = "Filip H.F. 'FiXato' Slagter <fixato [at] gmail [dot] com>"
SCRIPT_VERSION = "0.8.2"
SCRIPT_LICENSE = "MIT"
SCRIPT_DESC    = "A common buffer for /list output."
SCRIPT_COMMAND = "listbuffer"

import_ok = True

try:
  import weechat
except ImportError:
  print("This script must be run under WeeChat.")
  import_ok = False

import re

lb_settings = (
  ("autofocus", "on", "Focus the listbuffer in the current window if it isn't already displayed by a window."),
  ("sort_order", "users", "Last used sort order for the channel list."),
  ("sort_inverted", "on", "Invert the sort order for the channel list."),
  ("modes_min_width", "8", "The minimum width used for modes in the channel list. If a channel has less modes than this amount, the column will be padded with spaces."),
  ("channel_min_width", "25", "The minimum width used for the channel name in the channel list. If a channelname is shorter than this amount, the column will be padded with spaces."),
  ("users_min_width", "8", "The minimum width used for the usercount in the channel list. If the usercount has less digits than this amount, the column will be padded with spaces."),
)
lb_buffer = None
lb_curline = 0
lb_channels = []
lb_network = None
lb_list_started = False
lb_current_sort = None
lb_sort_inverted = False
lb_sort_options = (
  'channel',
  'users',
  'modes',
  'topic',
)

#                              server numeric Nick Chan  Users     Modes    Topic
lb_channel_list_expression = '(:\S+) (\d{3}) (\S+) (\S+) (\d+) :(\[(.*?)\] )?(.*)'

# Create listbuffer.
def lb_create_buffer():
  global lb_buffer, lb_curline

  if not lb_buffer:
    lb_buffer = weechat.buffer_new("listbuffer", "lb_input_cb", \
                "", "lb_close_cb", "")
    lb_set_buffer_title()
    # Sets notify to 0 as this buffer does not need to be in hotlist.
    weechat.buffer_set(lb_buffer, "notify", "0")
    weechat.buffer_set(lb_buffer, "nicklist", "0")
    weechat.buffer_set(lb_buffer, "type", "free")
    weechat.buffer_set(lb_buffer, "key_bind_ctrl-L", "/listbuffer **refresh")
    weechat.buffer_set(lb_buffer, "key_bind_meta2-A", "/listbuffer **up")
    weechat.buffer_set(lb_buffer, "key_bind_meta2-B", "/listbuffer **down")
    weechat.buffer_set(lb_buffer, "key_bind_meta2-1~", "/listbuffer **scroll_top")
    weechat.buffer_set(lb_buffer, "key_bind_meta2-4~", "/listbuffer **scroll_bottom")
    weechat.buffer_set(lb_buffer, "key_bind_meta-ctrl-J", "/listbuffer **enter")
    weechat.buffer_set(lb_buffer, "key_bind_meta-ctrl-M", "/listbuffer **enter")
    weechat.buffer_set(lb_buffer, "key_bind_meta->", "/listbuffer **sort_next")
    weechat.buffer_set(lb_buffer, "key_bind_meta-<", "/listbuffer **sort_previous")
    weechat.buffer_set(lb_buffer, "key_bind_meta-/", "/listbuffer **sort_invert")
    lb_curline = 0
  if weechat.config_get_plugin("autofocus") == "on":
    if not weechat.window_search_with_buffer(lb_buffer):
      weechat.command("", "/buffer " + weechat.buffer_get_string(lb_buffer,"name"))

def lb_set_buffer_title():
  global lb_buffer, lb_current_sort
  ascdesc = '(v)' if lb_sort_inverted else '(^)'
  weechat.buffer_set(lb_buffer, "title", lb_line_format({
    'channel': 'Channel name%s' % (ascdesc if lb_current_sort == 'channel' else ''),
    'users': 'Users%s' % (ascdesc if lb_current_sort == 'users' else ''),
    'modes': 'Modes%s' % (ascdesc if lb_current_sort == 'modes' else ''),
    'topic': 'Topic%s' % (ascdesc if lb_current_sort == 'topic' else ''),
    'nomodes': None,
  }))

def lb_list_start(data, signal, message):
  lb_initialise_list

  return weechat.WEECHAT_RC_OK

def lb_initialise_list(signal):
  global lb_channels, lb_network, lb_list_started

  lb_create_buffer()
  lb_channels = []
  lb_network = signal.split(',')[0]
  lb_list_started = True
  return


def lb_list_chan(data, signal, message):
  global lb_channels, lb_buffer, lb_list_started

  # Work-around for IRCds which don't send 321 Numeric (/List start)
  if not lb_list_started:
    lb_initialise_list(signal)

  for chan_data in re.findall(lb_channel_list_expression,message):
    lb_channels.append({
      'server':  chan_data[0][1:-1],
      'numeric': chan_data[1],
      'nick':    chan_data[2],
      'channel': chan_data[3],
      'users':   chan_data[4],
      'nomodes': chan_data[5] == '',
      'modes':   chan_data[6],
      'topic':   weechat.hook_modifier_exec("irc_color_decode", "1", chan_data[7])
    })
  return weechat.WEECHAT_RC_OK

def lb_list_end(data, signal, message):
  global lb_list_started

  # Work-around for IRCds which don't send 321 Numeric (/List start)
  if not lb_list_started:
    lb_initialise_list(signal)

  lb_list_started = False
  if lb_current_sort:
    lb_sort()
  lb_refresh()
  return weechat.WEECHAT_RC_OK

def keyEvent (data, buffer, args):
  global lb_options
  lb_options[args]()

def lb_input_cb(data, buffer, input_data):
  global lb_options, lb_curline
  lb_options[input_data]()
  return weechat.WEECHAT_RC_OK

def lb_refresh():
  global lb_channels, lb_buffer
  weechat.buffer_clear(lb_buffer)

  y = 0
  for list_data in lb_channels:
    lb_refresh_line(y)
    y += 1
  return

def lb_refresh_line(y):
  global lb_buffer, lb_curline, lb_channels
  if y >= 0 and y < len(lb_channels):
    formatted_line = lb_line_format(lb_channels[y], y == lb_curline)
    weechat.prnt_y(lb_buffer, y, formatted_line)

def lb_refresh_curline():
  global lb_curline
  lb_refresh_line(lb_curline-1)
  lb_refresh_line(lb_curline)
  lb_refresh_line(lb_curline+1)
  return

def lb_line_format(list_data,curr=False):
  str = ""
  if (curr):
    str += weechat.color("yellow,red")
  channel_text = list_data['channel'].ljust(int(weechat.config_get_plugin('channel_min_width')))
  users_text = "(%s)" % list_data['users']
  padded_users_text = users_text.rjust(int(weechat.config_get_plugin('users_min_width')) + 2)
  str += "%s%s %s " % (weechat.color("bold"), channel_text, padded_users_text)
  if not list_data['nomodes']:
    modes = "[%s]" % list_data['modes']
  else:
    modes = "[]"
  str += "%s: " % modes.rjust(int(weechat.config_get_plugin('modes_min_width')) + 2)
  str += "%s" % list_data['topic']
  return str

def lb_line_up():
  global lb_curline
  if lb_curline <= 0:
    return
  lb_curline -= 1
  lb_refresh_curline()
  lb_check_outside_window()
  return

def lb_line_down():
  global lb_curline, lb_channels
  if lb_curline+1 >= len(lb_channels):
    return
  lb_curline += 1
  lb_refresh_curline()
  lb_check_outside_window()
  return

def lb_line_run():
  global lb_channels, lb_curline, lb_network
  buff = weechat.info_get("irc_buffer", lb_network)
  channel = lb_channels[lb_curline]['channel']
  command = "/join %s" % channel
  weechat.command(buff, command)
  return

def lb_line_select():
  return

def lb_scroll_top():
  global lb_curline
  old_y = lb_curline
  lb_curline = 0
  lb_refresh_curline()
  lb_refresh_line(old_y)
  weechat.command(lb_buffer, "/window scroll_top")
  return

def lb_scroll_bottom():
  global lb_curline, lb_channels
  old_y = lb_curline
  lb_curline = len(lb_channels)-1
  lb_refresh_curline()
  lb_refresh_line(old_y)
  weechat.command(lb_buffer, "/window scroll_bottom")
  return

def lb_check_outside_window():
  global lb_buffer, lb_curline
  if (lb_buffer):
    infolist = weechat.infolist_get("window", "", "current")
    if (weechat.infolist_next(infolist)):
      start_line_y = weechat.infolist_integer(infolist, "start_line_y")
      chat_height = weechat.infolist_integer(infolist, "chat_height")
      if(start_line_y > lb_curline):
        weechat.command(lb_buffer, "/window scroll -%i" %(start_line_y - lb_curline))
      elif(start_line_y <= lb_curline - chat_height):
        weechat.command(lb_buffer, "/window scroll +%i"%(lb_curline - start_line_y - chat_height + 1))
    weechat.infolist_free(infolist)

def lb_sort_next():
  global lb_current_sort, lb_sort_options
  if lb_current_sort:
    new_index = lb_sort_options.index(lb_current_sort) + 1
  else:
    new_index = 0

  if len(lb_sort_options) <= new_index:
    new_index = 0

  lb_set_current_sort_order(lb_sort_options[new_index])
  lb_sort()

def lb_set_current_sort_order(value):
  global lb_current_sort
  lb_current_sort = value
  weechat.config_set_plugin('sort_order', lb_current_sort)

def lb_set_invert_sort_order(value):
  global lb_sort_inverted
  lb_sort_inverted = value
  weechat.config_set_plugin('sort_inverted', ('on' if lb_sort_inverted else 'off'))

def lb_sort_previous():
  global lb_current_sort, lb_sort_options
  if lb_current_sort:
    new_index = lb_sort_options.index(lb_current_sort) - 1
  else:
    new_index = 0

  if new_index < 0:
    new_index = len(lb_sort_options) - 1

  lb_set_current_sort_order(lb_sort_options[new_index])
  lb_sort()

def lb_sort(sort_key=None):
  global lb_channels, lb_current_sort, lb_sort_inverted
  if sort_key:
    lb_set_current_sort_order(sort_key)
  if lb_current_sort == 'users':
    lb_channels = sorted(lb_channels, key=lambda chan_data: int(chan_data[lb_current_sort]))
  else:
    lb_channels = sorted(lb_channels, key=lambda chan_data: chan_data[lb_current_sort])
  if lb_sort_inverted:
    lb_channels.reverse()
  lb_set_buffer_title()
  lb_refresh()

def lb_sort_invert():
  global lb_current_sort, lb_sort_inverted
  if lb_current_sort:
    lb_set_invert_sort_order(not lb_sort_inverted)
    lb_sort()

def lb_close_cb(*kwargs):
  """ A callback for buffer closing. """
  global lb_buffer

  lb_buffer = None
  return weechat.WEECHAT_RC_OK

lb_options = {
  'refresh'     : lb_refresh,
  'up'          : lb_line_up,
  'down'        : lb_line_down,
  'enter'       : lb_line_run,
  'space'       : lb_line_select,
  'scroll_top'  : lb_scroll_top,
  'scroll_bottom': lb_scroll_bottom,
  'sort_next'   : lb_sort_next,
  'sort_previous': lb_sort_previous,
  'sort_invert': lb_sort_invert
}

def lb_command_main(data, buffer, args):
  if args[0:2] == "**":
    keyEvent(data, buffer, args[2:])
  return weechat.WEECHAT_RC_OK

def lb_set_default_settings():
  global lb_settings
  # Set default settings
  for option, default_value, description in lb_settings:
     if not weechat.config_is_set_plugin(option):
         weechat.config_set_plugin(option, default_value)
         version = weechat.info_get("version_number", "") or 0
         if int(version) >= 0x00030500:
             weechat.config_set_desc_plugin(option, description)

def lb_reset_stored_sort_order():
  global lb_current_sort, lb_sort_inverted
  lb_current_sort = weechat.config_get_plugin('sort_order')
  lb_sort_inverted = (True if weechat.config_get_plugin('sort_inverted') == 'on' else False)

if __name__ == "__main__" and import_ok:
  if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION,
                      SCRIPT_LICENSE, SCRIPT_DESC, "lb_close_cb", ""):
    lb_set_default_settings()
    lb_reset_stored_sort_order()
    lb_buffer = weechat.buffer_search("python", "listbuffer")

    weechat.hook_signal("*,irc_in_321", "lb_list_start", "")
    weechat.hook_signal("*,irc_in_322", "lb_list_chan", "")
    weechat.hook_signal("*,irc_in_323", "lb_list_end", "")
    weechat.hook_command(SCRIPT_COMMAND,
                          "List Buffer",
                          "", "", "",
                          "lb_command_main", "")



import weechat
import diff_match_patch
import re

weechat.register('topicdiff_alt', 'Juerd <#####@juerd.nl>', '1.01', 'PD', "Announce topic with changes highlighted", '', '')

def topic(data, tags, msg):
    server = tags.split(",")[0]

    match = re.search(r':(\S+)\s+TOPIC\s+(\S+)\s+:(.*)', msg)

    if not match:
        return weechat.WEECHAT_RC_ERROR

    usermask, channel, newtopic = match.groups()
    nick, host = usermask.split("!", 1)

    buffer = weechat.buffer_search("irc", server + "." + channel)
    weechat.prnt("", server + "." + channel)

    if not buffer:
        return weechat.WEECHAT_RC_ERROR

    oldtopic = weechat.buffer_get_string(buffer, "title")
    if oldtopic == None:
        oldtopic = ""

    dmp = diff_match_patch.diff_match_patch()
    diff = dmp.diff_main(oldtopic, newtopic)
    dmp.diff_cleanupEfficiency(diff)

    topic = ""

    color_reset = weechat.color("reset")
    color_ins = weechat.color(weechat.config_get_plugin("color_ins"))
    color_del = weechat.color(weechat.config_get_plugin("color_del"))

    for chunk in diff:
        changed, text = chunk

        topic += "%s%s%s" % (
            # 0 (unchanged), 1 (added), -1 (removed)
            ["", color_ins, color_del][changed],
            text,
            ["", color_reset, color_reset][changed]
        )

    weechat.prnt_date_tags(buffer, 0, "irc_topicdiff",
        "%s%s%s%s has changed topic for %s%s%s: %s" % (
        weechat.prefix("network"),
        weechat.color(weechat.info_get("irc_nick_color", nick)) \
            if weechat.config_boolean("irc.look.color_nicks_in_server_messages") \
            else weechat.color("chat_nick"),
        nick,
        color_reset,
        weechat.color("chat_channel"),
        channel,
        color_reset,
        topic
    ))

    return weechat.WEECHAT_RC_OK

weechat.hook_signal("*,irc_in_topic", "topic", "")

if not weechat.config_is_set_plugin("color_ins"):
    weechat.config_set_plugin("color_ins", "lightcyan")

if not weechat.config_is_set_plugin("color_del"):
    weechat.config_set_plugin("color_del", "darkgray")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012 Sebastien Helleu <flashcode@flashtux.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# Samegame for WeeChat (http://en.wikipedia.org/wiki/SameGame).
# (requires WeeChat >= 0.3.6)
#
# History:
#
# 2012-03-16, Sebastien Helleu <flashcode@flashtux.org>:
#     version 0.2: add undo key and bonus +1000 when all blocks are removed
# 2012-03-16, Sebastien Helleu <flashcode@flashtux.org>:
#     version 0.1: initial release
#

SCRIPT_NAME    = 'samegame'
SCRIPT_AUTHOR  = 'Sebastien Helleu <flashcode@flashtux.org>'
SCRIPT_VERSION = '0.2'
SCRIPT_LICENSE = 'GPL3'
SCRIPT_DESC    = 'Samegame'

SCRIPT_COMMAND = 'samegame'

import_ok = True

try:
    import weechat
except ImportError:
    print('This script must be run under WeeChat.')
    print('Get WeeChat now at: http://www.weechat.org/')
    import_ok = False

try:
    import random, copy
except ImportError as message:
    print('Missing package(s) for %s: %s' % (SCRIPT_NAME, message))
    import_ok = False

samegame = {
    'buffer'    : '',
    'board'     : [],
    'sizes'     : ((15, 10), (25, 17)),
    'size'      : (15, 10),
    'zoom'      : 1,
    'colors'    : [],
    'numcolors' : 3,
    'score'     : 0,
    'end'       : '',
    'timer'     : '',
    'board_undo': None,
    'score_undo': 0,
}

# script options
samegame_settings_default = {
    'colors'   : ['blue,red,green,yellow,magenta,cyan', 'comma-separated list of 6 colors for blocks'],
    'numcolors': ['3',                                  'number of colors to use for blocks (3-6)'],
    'zoom'     : ['',                                   'zoom for board (0-N, empty means automatic zoom according to size of window)'],
    'speed'    : ['40',                                 'speed of animation when blocks are falling and columns are trimmed away (0 = immediate, 500 = slow animation'],
}
samegame_settings = {}

# mouse keys
samegame_mouse_keys = { '@chat(python.samegame):button1': '/window ${_window_number};hsignal:samegame_mouse' }

def samegame_display(clear=False):
    """Display status and board."""
    global samegame
    if not samegame['buffer']:
        return
    if clear:
        weechat.buffer_clear(samegame['buffer'])
    spaces = ' ' * ((samegame['zoom'] + 1) * 2)

    # display status
    str_status = 'Board: %s%dx%d%s    Colors: %s%d%s    Score: %s%d' % (weechat.color('white'), samegame['size'][0], samegame['size'][1],
                                                                        weechat.color('chat'),
                                                                        weechat.color('white'), samegame['numcolors'],
                                                                        weechat.color('chat'),
                                                                        weechat.color('white'), samegame['score'])
    str_end = '%s%s' % (weechat.color('white'), samegame['end'])
    weechat.prnt_y(samegame['buffer'], 0, '%s    %s' % (str_status, str_end))

    # display board
    weechat.prnt_y(samegame['buffer'], 1, '%s┌%s┐' % (weechat.color('chat'), '─' * (samegame['size'][0] * ((samegame['zoom'] + 1) * 2))))
    for y, line in enumerate(samegame['board']):
        str_line = '│'
        for color in line:
            if color < 0:
                str_color = 'default'
            else:
                str_color = samegame['colors'][color]
            str_line += '%s%s' % (weechat.color(',%s' % str_color), spaces)
        str_line += '%s│' % weechat.color('chat')
        for i in range (0, samegame['zoom'] + 1):
            weechat.prnt_y(samegame['buffer'], 2 + (y * (samegame['zoom'] + 1)) + i, str_line)
    weechat.prnt_y(samegame['buffer'], 1 + (samegame['size'][1] * (samegame['zoom'] + 1)) + 1,
                   '%s└%s┘' % (weechat.color('chat'), '─' * (samegame['size'][0] * ((samegame['zoom'] + 1) * 2))))

def samegame_adjust_zoom():
    """Choose zoom according to size of window."""
    global samegame, samegame_settings
    samegame['zoom'] = -1
    if samegame_settings['zoom']:
        try:
            samegame['zoom'] = int(samegame_settings['zoom'])
        except:
            samegame['zoom'] = -1
    if samegame['zoom'] < 0:
        width = weechat.window_get_integer(weechat.current_window(), 'win_chat_width')
        height = weechat.window_get_integer(weechat.current_window(), 'win_chat_height')
        for i in range(10, -1, -1):
            if width >= samegame['size'][0] * ((i + 1) * 2) + 2 and height >= (samegame['size'][1] * (i + 1)) + 3:
                samegame['zoom'] = i
                break
    if samegame['zoom'] < 0:
        samegame['zoom'] = 0

def samegame_set_colors():
    """Set list of colors using settings."""
    global samegame, samegame_settings, samegame_settings_default
    samegame['colors'] = samegame_settings['colors'].split(',')
    if len(samegame['colors']) != 6:
        weechat.prnt('', '%ssamegame: invalid colors (list must have 6 colors)' % weechat.prefix('error'))
        samegame['colors'] = samegame_settings_default['colors'][0].split(',')

def samegame_config_cb(data, option, value):
    """Called when a script option is changed."""
    global samegame, samegame_settings
    pos = option.rfind('.')
    if pos > 0:
        name = option[pos+1:]
        if name in samegame_settings:
            samegame_settings[name] = value
            if name == 'colors':
                samegame_set_colors()
            elif name == 'numcolors':
                try:
                    samegame['numcolors'] = int(value)
                except:
                    pass
            elif name == 'zoom':
                samegame_adjust_zoom()
    samegame_display()
    return weechat.WEECHAT_RC_OK

def samegame_new_game():
    """Create a new game: initialize board and some variables."""
    global samegame
    samegame['board'] = []
    for y in range(0, samegame['size'][1]):
        line = []
        for x in range(0, samegame['size'][0]):
            line.append(random.randint(0, samegame['numcolors'] - 1))
        samegame['board'].append(line)
    samegame['score'] = 0
    samegame['end'] = ''
    samegame_display()

def samegame_change_size(add):
    """Change size of board."""
    global samegame
    keys = sorted(samegame['sizes'])
    index = keys.index(samegame['size']) + add
    if index >= 0 and index < len(keys):
        samegame['size'] = keys[index]
        weechat.buffer_clear(samegame['buffer'])
        samegame_adjust_zoom()
        samegame_new_game()

def samegame_input_buffer(data, buffer, input):
    """Input data in samegame buffer."""
    global samegame
    if input:
        args = input.split(' ')
        if args[0] in ('n', 'new'):
            samegame_new_game()
        elif args[0] in ('q', 'quit'):
            weechat.buffer_close(samegame['buffer'])
        elif args[0] == '+':
            samegame_change_size(+1)
        elif args[0] == '-':
            samegame_change_size(-1)
        elif args[0].isdigit():
            numcolors = int(args[0])
            if numcolors >= 3 and numcolors <= 6:
                samegame['numcolors'] = numcolors
                samegame_new_game()
    return weechat.WEECHAT_RC_OK

def samegame_close_buffer(data, buffer):
    """Called when samegame buffer is closed."""
    global samegame
    if samegame['timer']:
        weechat.unhook(samegame['timer'])
        samegame['timer'] = ''
    samegame['buffer'] = ''
    return weechat.WEECHAT_RC_OK

def samegame_init():
    """Init samegame: create buffer, adjust zoom, new game."""
    global samegame, samegame_settings
    if samegame['buffer']:
        return
    samegame['buffer'] = weechat.buffer_search('python', 'samegame')
    if not samegame['buffer']:
        samegame['buffer'] = weechat.buffer_new('samegame', 'samegame_input_buffer', '', 'samegame_close_buffer', '')
        if samegame['buffer']:
            weechat.buffer_set(samegame['buffer'], 'type', 'free')
            weechat.buffer_set(samegame['buffer'], 'title',
                               'Samegame | mouse: play, alt-n: new game, alt-+/-: adjust board zoom, alt-u: undo | '
                               'Command line: (n)ew, +/-: change size, (q)uit, 3-6: number of colors')
            weechat.buffer_set(samegame['buffer'], 'key_bind_meta-n',  '/samegame new')
            weechat.buffer_set(samegame['buffer'], 'key_bind_meta-+',  '/samegame zoom')
            weechat.buffer_set(samegame['buffer'], 'key_bind_meta--',  '/samegame dezoom')
            weechat.buffer_set(samegame['buffer'], 'key_bind_meta-u',  '/samegame undo')
    try:
        samegame['numcolors'] = int(samegame_settings['numcolors'])
    except:
        pass
    if samegame['numcolors'] < 3:
        samegame['numcolors'] = 3
    if samegame['numcolors'] > 6:
        samegame['numcolors'] = 6
    if samegame['buffer']:
        samegame_adjust_zoom()
        samegame_new_game()

def samegame_play_xy(board, x, y):
    """Play at (x,y) and return number of blocks removed."""
    color = board[y][x]
    count = 1
    board[y][x] = -1
    if y > 0 and board[y-1][x] == color:
        count += samegame_play_xy(board, x, y - 1)
    if y < samegame['size'][1] - 1 and board[y+1][x] == color:
        count += samegame_play_xy(board, x, y + 1)
    if x > 0 and board[y][x-1] == color:
        count += samegame_play_xy(board, x - 1, y)
    if x < samegame['size'][0] - 1 and board[y][x+1] == color:
        count += samegame_play_xy(board, x + 1, y)
    return count

def samegame_column_is_empty(x):
    """Return True if a column is empty."""
    global samegame
    count = 0
    for y in range(0, samegame['size'][1]):
        if samegame['board'][y][x] < 0:
            count += 1
    return (count == samegame['size'][1])

def samegame_collapse_blocks():
    """Collapse blocks."""
    global samegame
    columns = []
    for x in range(0, samegame['size'][0]):
        for y in range(0, samegame['size'][1] - 1):
            if samegame['board'][y][x] >= 0 and samegame['board'][y+1][x] < 0:
                columns.append(x)
                break
    if columns:
        for x in columns:
            for y in range(samegame['size'][1] - 1, 0, -1):
                if samegame['board'][y][x] < 0 and samegame['board'][y-1][x] >= 0:
                    samegame['board'][y][x] = samegame['board'][y-1][x]
                    samegame['board'][y-1][x] = -1
        return True
    for x in range(0, samegame['size'][0] - 1):
        if samegame_column_is_empty(x) and not samegame_column_is_empty(x + 1):
            for x2 in range(x, samegame['size'][0] - 1):
                for y in range(0, samegame['size'][1]):
                    samegame['board'][y][x2] = samegame['board'][y][x2+1]
            for y in range(0, samegame['size'][1]):
                samegame['board'][y][samegame['size'][0]-1] = -1
            return True
    return False

def samegame_count_color(board, color):
    """Count number of times a color is used in board."""
    count = 0
    for line in board:
        count += line.count(color)
    return count

def samegame_check_end():
    """Check if the game has ended (play is not possible with remaining blocks)."""
    global samegame
    board = copy.deepcopy(samegame['board'])
    for x in range(0, samegame['size'][0]):
        for y in range(0, samegame['size'][1]):
            if board[y][x] >= 0:
                if samegame_play_xy(board, x, y) >= 2:
                    return False
    samegame['end'] = 'End of game!'
    blocks_remaining = (samegame['size'][0] * samegame['size'][1]) - samegame_count_color(samegame['board'], -1)
    if blocks_remaining == 0:
        samegame['end'] += '  ** CONGRATS! **'
        samegame['score'] += 1000
    else:
        samegame['end'] += '  (%d blocks remaining)' % blocks_remaining
    return True

def samegame_timer_cb(data, remaining_calls):
    """Timer for animation (blocks falling)."""
    global samegame
    if samegame_collapse_blocks():
        samegame_display()
    else:
        weechat.unhook(samegame['timer'])
        samegame['timer'] = ''
        if samegame_check_end():
            samegame_display()
    return weechat.WEECHAT_RC_OK

def samegame_play(x, y):
    """Play at (x,y), and check if game has ended."""
    global samegame, samegame_settings
    if samegame['board'][y][x] < 0:
        return
    board = copy.deepcopy(samegame['board'])
    count = samegame_play_xy(board, x, y)
    if count < 2:
        return
    samegame['board_undo'] = copy.deepcopy(samegame['board'])
    samegame['score_undo'] = samegame['score']
    count = samegame_play_xy(samegame['board'], x, y)
    samegame['score'] += (count - 1) ** 2
    delay = 50
    try:
        delay = int(samegame_settings['speed'])
    except:
        delay = 50
    if delay < 0:
        delay = 0
    elif delay > 500:
        delay = 500
    if delay == 0:
        while samegame_collapse_blocks():
            pass
        samegame_check_end()
    else:
        samegame['timer'] = weechat.hook_timer(delay, 0, 0, 'samegame_timer_cb', '')
    samegame_display()
    return

def samegame_cmd_cb(data, buffer, args):
    """The /samegame command."""
    global samegame
    samegame_init()
    if samegame['buffer']:
        weechat.buffer_set(samegame['buffer'], 'display', '1')
    if args == 'new':
        samegame_new_game()
    elif args == 'zoom':
        samegame['zoom'] += 1
        samegame_display(True)
    elif args == 'dezoom':
        if samegame['zoom'] > 0:
            samegame['zoom'] -= 1
            samegame_display(True)
    if not samegame['end']:
        if args == 'undo':
            if samegame['board_undo']:
                samegame['board'] = copy.deepcopy(samegame['board_undo'])
                samegame['board_undo'] = None
                samegame['score'] = samegame['score_undo']
                samegame['score_undo'] = 0
                samegame_display()
    return weechat.WEECHAT_RC_OK

def samegame_mouse_cb(data, hsignal, hashtable):
    """Mouse callback."""
    global samegame
    if not samegame['end'] and not samegame['timer']:
        x = int(hashtable.get('_chat_line_x', '-1'))
        y = int(hashtable.get('_chat_line_y', '-1'))
        if x >= 0 and y >= 0:
            color = -1
            if y >= 2:
                x = x // ((samegame['zoom'] + 1) * 2)
                y = (y - 2) // (samegame['zoom'] + 1)
                if y < samegame['size'][1] and x < samegame['size'][0]:
                    color = samegame['board'][y][x]
            if color >= 0:
                samegame_play(x, y)
    return weechat.WEECHAT_RC_OK

if __name__ == '__main__' and import_ok:
    if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE,
                        SCRIPT_DESC, '', ''):
        # set default settings
        for option, value in samegame_settings_default.items():
            if weechat.config_is_set_plugin(option):
                samegame_settings[option] = weechat.config_get_plugin(option)
            else:
                weechat.config_set_plugin(option, value[0])
                samegame_settings[option] = value[0]
            weechat.config_set_desc_plugin(option, '%s (default: "%s")' % (value[1], value[0]))
        samegame_set_colors()

        # mouse support
        weechat.key_bind('mouse', samegame_mouse_keys)
        weechat.hook_hsignal('samegame_mouse', 'samegame_mouse_cb', '')

        # detect config changes
        weechat.hook_config('plugins.var.python.%s.*' % SCRIPT_NAME, 'samegame_config_cb', '')

        # add command
        weechat.hook_command(SCRIPT_COMMAND, 'Samegame.', '',
                             'Instructions:\n'
                             '- click on a group of adjoining blocks of the same color to remove them from the screen\n'
                             '- blocks that are no longer supported will fall down, and a column without any blocks will be trimmed away by other columns sliding to the left\n'
                             '- your score is increased by (N-1)², where N is the number of blocks removed by your click\n'
                             '- the game ends when you can not play any more.',
                             '', 'samegame_cmd_cb', '')

        # if buffer already exists (after /upgrade), init samegame
        if weechat.buffer_search('python', 'samegame'):
            samegame_init()



# -*- coding: utf-8 -*-

# Copyright (c) 2017-2019 oakkitten <jatjasjem@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

'''
Why yet another url script? Well, this is what sets this one apart from others:
  * always visible hints for urls (by default looks like ¹http://this)
  * the hints change as the urls appear. hint ¹ always points to the last url, ² to the second last, etc
  * you can open these with keyboard shortcuts — just one key press to open an url!
  * it can be made to work even when your weechat is on a remote machine through your OS or terminal emulator
  * a ready recipe for opening urls from your PuTTY!

So, this script prepends tiny digits to links, ¹ to the latest url, ² to the second latest, etc. \\
Also, it can put these urls in the window title, which you can grab using your OS automation, or your terminal emulator. \\
This is an example script in AutoHotkey that works with PuTTY (just install ahk, save this to file.ahk and run it):

  #IfWinActive, ahk_class PuTTY
    F1::
    F2::
    F3::
    F4::
    F5::
      WinGetTitle, title
      if (!RegExMatch(title, "[^""]+""urls: (.+)""", out))
        Return
      url := StrSplit(out1, " ")[SubStr(A_ThisHotkey, 2)]
      If (RegExMatch(url, "^(?:http|www)"))
        Run, %url%
    Return
  #IfWinActive

You can use command /url_hint_replace that replaces {url1}, {url2}, etc with according urls and then \\
executes the result. For example, the following will open url 1 in your default browser:
  /url_hint_replace /exec -bg xdg-open {url1}

or in elinks a new tmux window:
  /url_hint_replace /exec -bg tmux new-window elinks {url1}

You can bind opening of url 1 to f1 and url 2 to f2 like this, for example (press meta-k, then f1. that prints `meta2-11~`):
  /alias add open_url /url_hint_replace /exec -bg tmux new-window elinks {url$1}
  /key bind meta2-11~ /open_url 1
  /key bind meta2-12~ /open_url 2

WARNING: Avoid passing urls to the shell as they can contain special characters. \\
If you must do that, consider setting the option safe_urls to "base64".

Configuration (plugins.var.python.url_hint.*):
  * max_lines: the maximum number of lines that contain urls to track ("10")
  * no_urls_title: title for buffers that don't contain urls ("weechat")
  * prefix: the beginning of the title when there are urls ("urls: ")
  * delimiter: what goes between the urls in the title (" ")
  * postfix: the end of the title when there are urls ("")
  * hints: comma-separated list of hints. evaluated, can contain colors ("⁰,¹,²,³,⁴,⁵,⁶,⁷,⁸,⁹")
  * update_title: whether the script should put urls into the title ("on")
  * safe_urls: whether the script will convert urls to their safe ascii equivalents. can be either "off", \\
    "on" for idna- & percent-encoding, or "base64" for utf-8 base64 encoding ("off")

Notes:
  * to avoid auto renaming tmux windows use :set allow-rename off
  * in PuTTyTray and possibly other clients, window titles are parsed using wrong charset (see bug #88). \\
    The option safe_urls must be set to non-"off" value to avoid issues

Limitations:
  * will not work with urls that have color codes inside them
  * will be somewhat useless in merged and zoomed buffers

Version history:
  0.8 (21 august 2019): fixed minor issues in url detections and improved code quality
  0.7 (4 august 2019): py3 compatibility
  0.6 (26 june 2017): renamed /url_hint to /url_hint_replace; /url_hint simply prints help now
  0.5 (22 june 2017): implemented base64 encoding of urls
  0.4 (18 june 2017): encode fewer characters for safe urls—helps with servers that don't follow the rfc
  0.3 (10 june 2017): don't fail when safe url encoding fails
  0.2 (4 june 2017): don't crash if a new buffer has the same pointer as the old one
  0.1 (30 may 2017): added an option to make safe urls
  0.0 (6 may 2017): initial release
'''

from __future__ import unicode_literals, print_function
import re
import sys
from base64 import b64encode

MYPY = False
if MYPY:
    # noinspection PyUnresolvedReferences
    from typing import Match, Pattern, Generator, List, Union, Optional, Any, Callable, Dict, Tuple

PY3 = sys.version_info[0] >= 3
if PY3:
    # noinspection PyUnresolvedReferences,PyCompatibility
    from urllib.parse import quote_from_bytes, unquote_to_bytes
    from_weechat = to_weechat = lambda s: s
else:
    # noinspection PyUnresolvedReferences
    from urllib import quote as quote_from_bytes, unquote as unquote_to_bytes
    from_weechat, to_weechat = lambda s: s.decode("utf-8"), lambda s: s.encode("utf-8")
    # noinspection PyUnresolvedReferences,PyShadowingBuiltins
    str = unicode

SCRIPT_NAME = "url_hint"
SCRIPT_VERSION = "0.8"
SCRIPT_DOC = re.sub(r" *\\\n *", " ", __doc__.strip())

# the following code constructs a simple but neat regular expression for detecting urls
# it's by no means perfect, but it will detect an url in quotes and parentheses, http iris,
# punycode, urls followed by punctuation and such
def construct_url_finder_pattern():
    ipv4_segment = r"(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)"
    ipv6_segment = r"[0-9A-Fa-f]{1,4}"

    # 00-1f     c0 control chars
    # 20        space
    # 21-2f     !"#$%&'()*+,-./
    # 30-39         0123456789
    # 3a-40     :;<=>?@
    # 41-5a         ABCDEFGHIJKLMNOPQRSTUVWXYZ
    # 5b-60     [\]^_`
    # 61-7a         abcdefghijklmnopqrstuvwxyz
    # 7b-7e     {|}~
    # 7f        del
    # 80-9f     c1 control chars
    # a0        nbsp
    good_ascii_char = r"[A-Za-z0-9]"
    bad_char = r"\x00-\x20\x7f-\xa0\ufff0-\uffff\s"
    good_char = r"[^{bad}]".format(bad=bad_char)
    good_host_char = r"[^\x00-\x2f\x3a-\x40\x5b-\x60\x7b-\xa0\ufff0-\uffff]"
    good_tld_char = r"[^\x00-\x40\x5b-\x60\x7b-\xa0\ufff0-\uffff]"

    host_segment = r"""{good_host}+(?:-+{good_host}+)*""".format(good_host=good_host_char)
    tld = r"(?:{good_tld}{{2,}}|xn--{good_ascii}+)".format(good_ascii=good_ascii_char, good_tld=good_tld_char)

    # language=PythonVerboseRegExp
    url = r"""
        # url must be preceded by a word boundary, or a weechat color (a control character followed by a digit)
        (?:(?<=\d)|\b)
    
        # http:// or www  =1=
        (https?://|www\.)
    
        # optional userinfo at  =2=
        (?:([^{bad}@]*)@)?
    
        # ip or host  =3=
        (
            # ipv4
            (?:(?:{s4}\.){{3}}{s4})
        |
            # ipv6 (no embedded ipv4 tho)
            \[
            (?:
                                              (?:{s6}:){{7}} {s6}
                |                          :: (?:{s6}:){{6}} {s6}
                | (?:               {s6})? :: (?:{s6}:){{5}} {s6}
                | (?:(?:{s6}:)?     {s6})? :: (?:{s6}:){{4}} {s6}
                | (?:(?:{s6}:){{,2}}{s6})? :: (?:{s6}:){{3}} {s6}
                | (?:(?:{s6}:){{,3}}{s6})? :: (?:{s6}:){{2}} {s6}
                | (?:(?:{s6}:){{,4}}{s6})? :: (?:{s6}:)      {s6}
                | (?:(?:{s6}:){{,5}}{s6})? ::                {s6}
                | (?:(?:{s6}:){{,6}}{s6})? ::
            )
            \]
        |
            # domain name (a.b.c.com)
            {host_segment}
            (?:\.{host_segment})*
            \.{tld}
            
            # fqdn dot, but only if followed by url-ish things
            (?:\.(?=/|:\d))?
        )
    
        # port?  =4=
        (:\d{{1,5}})?
    
        # / & the rest  =5=
        (
            /
            # hello(world) in "hello(world))"
            (?:
                [^{bad}(]*
                \(
                [^{bad})]+
                \)
            )*
            # any string (non-greedy!)
            {good}*?
        )?
    
        # url must be directly followed by:
        (?=
            # some possible punctuation
            # AND space or end of string
            [\]>,.)!?:'"”@]*
            (?:[{bad}]|$)
        )
        """
    url = url.format(s4=ipv4_segment, s6=ipv6_segment, bad=bad_char, good=good_char, host_segment=host_segment, tld=tld)
    return re.compile(url, re.U | re.X | re.I)
RE_URL = construct_url_finder_pattern()

H_BUFFER, H_LINES, H_LINE, H_LINE_DATA, H_GUI_BUFFERS = None, None, None, None, None    # make mypy happy

###############################################################################
###############################################################################
###############################################################################

# noinspection PyPep8Naming
class lazy(object):
    def __init__(self, fget):
        self.fget = fget

    def __get__(self, obj, cls):
        value = self.fget(obj)
        setattr(obj, self.fget.__name__, value)
        return value

class Url(object):
    """
    an object that represents an url

    exact (str): exact url, may contain non-ascii characters
    safe (str): safe url using punycode and percent-escaped characters
    base64 (str): safe url, encoded with utf-8 base64
    url (str): exact or safe or base64, depending on current settings
    """
    def __init__(self, match):                      # type: (Match) -> None
        self._match = match
        self.exact = match.group(0)

    @lazy
    def safe(self):                                 # type: () -> str
        prefix, userinfo, ip_or_host, c_port, rest = self._match.groups()
        safe = prefix
        if userinfo: safe += q(userinfo, safe=SAFE_USERINFO) + "@"
        safe += ip_or_host.encode("idna").decode("utf-8")
        if c_port: safe += c_port
        if rest:
            end = ""
            if "#" in rest:
                rest, fragment = rest.split("#", 1)
                end = "#" + q(fragment, safe=SAFE_FRAGMENT)
            if "?" in rest:
                rest, query = rest.split("?", 1)
                end = "?" + q(query, safe=SAFE_QUERY) + end
            safe += "/".join(q(segment, safe=SAFE_PATH) for segment in rest.split("/"))
            safe += end
        return safe

    @lazy
    def base64(self):                               # type: () -> str
        return b64encode(self.exact.encode("utf-8")).decode("utf-8")

    @property
    def url(self):                                  # type: () -> str
        return getattr(self, C[SAFE])

# these are used to encode the url in a safe manner. basically it's url normalization, but it will not fail while
# encoding invalid urls and it changes as little as possible
# the following are direct quotes from https://tools.ietf.org/html/rfc3986
#     unreserved    = ALPHA / DIGIT / "-" / "." / "_" / "~"
#     sub-delims    = "!" / "$" / "&" / "'" / "(" / ")" / "*" / "+" / "," / ";" / "="
#     pct-encoded   = "%" HEXDIG HEXDIG
#     pchar         = unreserved / pct-encoded / sub-delims / ":" / "@"
#     userinfo      = *( unreserved / pct-encoded / sub-delims / ":" )
#     segment       = *pchar
#     query         = *( pchar / "/" / "?" )
#     fragment      = *( pchar / "/" / "?" )
SUB_DELIMS = b"!$&'()*+,;="
SAFE_USERINFO = SUB_DELIMS + b":"
SAFE_PATH = SUB_DELIMS + b":@"
SAFE_FRAGMENT = SAFE_QUERY = SUB_DELIMS + b":@" + b"/?"

# this uses bytes as urls do not necessarily contain unicode in them. for instance,
# the string `%FF?%FE#%FF` unquoted to unicode becomes `�?�#�` (that's 3 U+FFFD replacement
# characters) instead of b'\xff?\xfe#\xff'.
def q(string, safe):                                # type: (str, bytes) -> str
    out = quote_from_bytes(unquote_to_bytes(string.encode("utf-8")), safe=safe)    # type: Any
    return out if PY3 else out.decode("utf-8")

def find_urls(string):                              # type: (str) -> Generator[Union[str, Url], None, None]
    last_end = 0
    for match in RE_URL.finditer(string):
        if not is_valid_url(match):
            continue
        start, end = match.span()
        yield string[last_end:start]
        yield Url(match)
        last_end = end
    yield string[last_end:]

# domain name (a.b.c.d.) can be a maximum of 253 characters, not counting the FQDN dot;
# each label (a, b, ...) can be a maximum of 63 characters long
def is_valid_url(match):                            # type: (Match) -> bool
    domain_name = match.group(3).rstrip(".")
    try:
        domain_name.encode("idna")
    except UnicodeError:                            # fired if label length is wrong
        return False
    if len(domain_name) > 253:
        return False
    return True

###############################################################################
###############################################################################
###############################################################################

class Line(object):
    """
    an object that represents a buffer line with urls

    pointer (str): a hex pointer to line object in weechat
    urls ([Url]): a list of urls in reverse order

    is_of_pointer(pointer): True if this line corresponds to given pointer
    redraw(index): redraw the line and assign new hints to urls, given index of last hint
    reset(): restore the line to its original condition
    """
    def __init__(self, pointer, message, parts):    # type: (str, str, List[Union[str, Url]]) -> None
        assert pointer
        self.pointer = pointer
        self.urls = parts[-2::-2]                   # type: List[Url] # type: ignore
        self._data = weechat.hdata_pointer(H_LINE, pointer, "data")
        self._original_message = message
        self._current_message = message
        self._parts = [part.exact if isinstance(part, Url) else part for part in parts]
        for i in range(1, len(self._parts) * 3 // 2, 3):
            self._parts.insert(i, None)

    def is_of_pointer(self, pointer):               # type: (str) -> bool
        return self.pointer == pointer and self._data == weechat.hdata_pointer(H_LINE, pointer, "data") and \
               self._current_message == from_weechat(weechat.hdata_string(H_LINE_DATA, self._data, "message"))

    def redraw(self, index):                        # type: (int) -> None
        self._parts[1::3] = (get_hint(i) for i in reversed(range(index, index + len(self.urls))))
        self._set_message("".join(self._parts))

    def reset(self):                                # type: () -> None
        self._set_message(self._original_message)

    def _set_message(self, message):                # type: (str) -> None
        self._current_message = message
        weechat.hdata_update(H_LINE_DATA, self._data, {"message": to_weechat(message)})

def get_hint(i):                                    # type: (int) -> str
    hints = C[HINTS]
    def do_it(j):
        d, m = divmod(j, len(hints))
        return do_it(d) + hints[m] if d else hints[m]
    return do_it(i)

###############################################################################
###############################################################################
###############################################################################

class Buffer(object):
    """
    an object that represents a buffer

    pointer (str): a hex pointer to buffer object in weechat
    urls ([Url]): a list of recent urls, in reverse order

    on_message(): called on every visible message on a buffer
    valid(): returns True if this buffer still exists in weechat and is safe to operate on
    redraw(full_reset): redraw all lines. if full_reset is True, restores all lines to their original conditions
    """
    def __init__(self, pointer):                    # type: (str) -> None
        assert pointer
        self.pointer = pointer
        self.urls = []                              # type: List[Url]
        self._lines = []                            # type: List[Line]      # (reverse order)

    def on_message(self, message):                  # type: (str) -> None
        parts = list(find_urls(message))            # parts = ["text", url, "text", url, ""]
        if len(parts) < 3:
            return
        line_pointer = self.get_last_line_pointer()
        if not line_pointer:
            return
        line = Line(line_pointer, message, parts)
        self._lines.insert(0, line)
        self.redraw(last_pointer=line.pointer)

    def redraw(self, last_pointer=None, full_reset=False):  # type: (Optional[str], bool) -> None
        if last_pointer is None:
            last_pointer = self.get_last_line_pointer()

        max_lines = 0 if full_reset else C[MAX_LINES]       # type: int
        lines, urls = [], []                                # type: List[Line], List[Url]
        walker = line_reverse_walker(last_pointer)
        for n, line in enumerate(self._lines):
            for pointer in walker:
                if line.is_of_pointer(pointer):
                    if n >= max_lines:
                        line.reset()
                    else:
                        line.redraw(len(urls) + 1)
                        urls += line.urls
                        lines.append(line)
                    break
            else:
                break
        self._lines, self.urls = lines, urls

    # return pointer to the last line, visible or not; None otherwise
    def get_last_line_pointer(self):                # type: () -> Optional[str]
        try:
            assert weechat.hdata_check_pointer(H_BUFFER, H_GUI_BUFFERS, self.pointer)
            own_lines = weechat.hdata_pointer(H_BUFFER, self.pointer, "own_lines"); assert own_lines
            last_line = weechat.hdata_pointer(H_LINES, own_lines, "last_line"); assert last_line
        except AssertionError:
            return None
        return last_line

def line_reverse_walker(pointer):                   # type: (Optional[str]) -> Generator[str, None, None]
    while pointer:
        yield pointer
        pointer = weechat.hdata_move(H_LINE, pointer, -1)

###############################################################################
###############################################################################
###############################################################################

class Buffers(dict):
    def __missing__(self, key):
        return self.setdefault(key, Buffer(key))
buffers = Buffers()
current_buffer = ""

# see the note at https://weechat.org/files/doc/devel/weechat_plugin_api.en.html#_hook_print
def on_print(_data, pointer, _date, _tags, displayed, _highlighted, _prefix, message):   # type: (str, str, str, str, Union[str, int], Union[str, int], str, str) -> int
    displayed = int(displayed)
    if displayed:
        buffers[pointer].on_message(from_weechat(message))
        if C[UPDATE_TITLE] and current_buffer == pointer:
            update_title()
    return weechat.WEECHAT_RC_OK

def on_buffer_switch(_data, _signal, pointer):      # type: (str, str, str) -> int
    global current_buffer
    if C[UPDATE_TITLE] and current_buffer != pointer:
        current_buffer = pointer
        update_title()
    return weechat.WEECHAT_RC_OK

def update_title():                                 # type: () -> None
    if not current_buffer: return
    urls = buffers[current_buffer].urls
    title = (C[PREFIX] + C[DELIMITER].join(url.url for url in urls) + C[POSTFIX]) if urls else C[NO_URLS_TITLE]     # type: ignore
    weechat.window_set_title(to_weechat(title))

RE_REP = re.compile(r"{url(\d+)}", re.I)

# simply print help
def url_hint(_data, _pointer, _command):            # type: (str, str, str) -> int
    weechat.prnt("", to_weechat(SCRIPT_DOC))
    return weechat.WEECHAT_RC_OK

def url_hint_replace(_data, pointer, command):      # type: (str, str, str) -> int
    urls = buffers[pointer].urls
    def get_url(match):
        try: return urls[int(match.group(1)) - 1].url
        except IndexError: raise IndexError("could not replace " + match.group(0))
    try: weechat.command(pointer, to_weechat(RE_REP.sub(get_url, from_weechat(command))))
    except IndexError as e: print_error(str(e))
    return weechat.WEECHAT_RC_OK

def print_error(text):                              # type: (str) -> None
    weechat.prnt("", to_weechat("%s%s: %s" % (weechat.prefix("error"), SCRIPT_NAME, text)))

def redraw_everything(full_reset):                  # type: (bool) -> int
    for pointer, buffer in buffers.items():
        buffer.redraw(full_reset=full_reset)
    update_title()
    return weechat.WEECHAT_RC_OK

def exit_function():                                # type: () -> int
    return redraw_everything(full_reset=True)

###############################################################################
###############################################################################
###############################################################################

def hints_from_string(x):                           # type: (str) -> Optional[List[str]]
    hints = from_weechat(weechat.string_eval_expression(to_weechat(x), "", "", "")).split(",")
    return None if len(hints) < 2 or any(not hint.strip() for hint in hints) else hints

def boolean_from_string(x):                         # type: (str) -> Optional[bool]
    return {"on": True, "off": False}.get(x.lower(), None)

def safe_from_string(x):                            # type: (str) -> Optional[str]
    return {"off": "exact", "on": "safe", "base64": "base64"}.get(x.lower(), None)

MAX_LINES = "max_lines"
NO_URLS_TITLE = "no_urls_title"
PREFIX, DELIMITER, POSTFIX = "prefix", "delimiter", "postfix"
UPDATE_TITLE = "update_title"
HINTS = "hints"
SAFE = "safe_urls"

# setting name: (default value, default value as stored in weechat, a method of converting/validating that
# returns None if invalid, description)
DEFAULT_CONFIG = {
    MAX_LINES: ("10", "the maximum number of lines that contain urls to track", lambda x: int(x) if x.isdigit() and int(x) > 0 else None),
    NO_URLS_TITLE: ("weechat", "title for buffers that don't contain urls", None),
    PREFIX: ("urls: ", "the beginning of the title when there are urls", None),
    DELIMITER: (" ", "what goes between the urls in the title", None),
    POSTFIX: ("", "the end of the title when there are urls", None),
    HINTS: ("⁰,¹,²,³,⁴,⁵,⁶,⁷,⁸,⁹", "comma-separated list of hints. evaluated, can contain colors", hints_from_string),
    UPDATE_TITLE: ("on", "whether the script should put urls into the title", boolean_from_string),
    SAFE: ("off", """whether the script will convert urls to their safe ascii equivalents. can be either "off", "on" for idna- & percent-encoding, or "base64" for utf-8 base64 encoding""", safe_from_string)
}                                                   # type: Dict[str, Tuple[str, str, Optional[Callable]]]

C = {}                                              # type: Dict[str, Any]

def load_config(*_):                                # type: (Any) -> int
    for name, (default, description, from_string) in DEFAULT_CONFIG.items():
        value = None
        if weechat.config_is_set_plugin(name):
            value = from_weechat(weechat.config_get_plugin(name))
            if from_string: value = from_string(value)
        if value is None:
            value = from_string(default) if from_string else default
            weechat.config_set_plugin(name, to_weechat(default))
            weechat.config_set_desc_plugin(name, to_weechat('%s (default: "%s")' % (description, default)))
        C[name] = value
    redraw_everything(full_reset=False)
    return weechat.WEECHAT_RC_OK

###############################################################################
###############################################################################
###############################################################################

def install():
    global H_BUFFER, H_LINES, H_LINE, H_LINE_DATA, H_GUI_BUFFERS
    if not weechat.register(SCRIPT_NAME, "oakkitten", SCRIPT_VERSION, "MIT",
                            "Display hints for urls and open them with keyboard shortcuts", "exit_function", ""):
        raise Exception("Could not register script")

    weechat_version = int(weechat.info_get('version_number', '') or 0)
    if weechat_version <= 0x00040000:
        raise Exception("Need Weechat 0.4 or higher")

    H_BUFFER = weechat.hdata_get("buffer")
    H_LINES = weechat.hdata_get("lines")
    H_LINE = weechat.hdata_get("line")
    H_LINE_DATA = weechat.hdata_get("line_data")
    H_GUI_BUFFERS = weechat.hdata_get_list(H_BUFFER, "gui_buffers")

    load_config()

    weechat.hook_print("", "", "", 0, "on_print", "")
    weechat.hook_signal("buffer_switch", "on_buffer_switch", "")
    weechat.hook_config("plugins.var.python." + SCRIPT_NAME + ".*", "load_config", "")

    if not weechat.hook_command("url_hint", to_weechat(SCRIPT_DOC), "", "", "", "url_hint", ""):
        print_error("could not hook command /url_hint")

    if not weechat.hook_command("url_hint_replace", """Replaces {url1} with url hinted with a 1, etc. Examples:

Open url 1 in your default browser:
  /url_hint_replace /exec -bg xdg-open {url1}

Open url 1 in elinks in a new tmux window:
  /url_hint_replace /exec -bg tmux new-window elinks {url1}

Bind opening of url 1 to F1 and url 2 to F2 (press meta-k, then f1. that prints "meta2-11~"):
  /alias add open_url /url_hint_replace /exec -bg tmux new-window elinks {url$1}
  /key bind meta2-11~ /open_url 1
  /key bind meta2-12~ /open_url 2""", "<command>", "", "", "url_hint_replace", ""):
        print_error("could not hook command /url_hint_replace")

###############################################################################
###############################################################################
###############################################################################

def run_tests():
    if sys.version_info >= (3, 5):
        print("running mypy…")
        import subprocess
        if subprocess.call(["mypy", __file__], shell=True) == 0:
            print("no errors found")
    else:
        print("skipping mypy as it's not available on python version < 3.5")

    # noinspection SpellCheckingInspection
    tests = (
        "foo",
        "http://",
        "http://#",
        "http:// fail.com",
        "http://.www..foo.bar/",
        "http://url.c",
        "http://url.co1/,",
        "http://2.2.2.256/foo ",
        "http://[3210:123z::]:80/bye#",
        "www.mail-.lv",
        "www.ma.-il.lv"
        "http://squirrel",                                                 # this is valid but we don't want it anyway,
        "wut://server.com",
        "http://ser$ver.com",
        "http://ser_ver.com",
        "http://www.abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijkl.com/",     # label too long
        "http://a.bcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcde.abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.com",    # domain name too long

        ("http://url.co\u00a0m/,", "http://url.co"),     # non-breaking space
        ("[http://[3ffe:2a00:100:7031::1]", "http://[3ffe:2a00:100:7031::1]"),
        ("http://[1080::8:800:200C:417A]/foo)", "http://[1080::8:800:200C:417A]/foo"),
        ("http://[FEDC:BA98:7654:3210:FEDC:BA98:7654:3210]:80/index.html", "http://[FEDC:BA98:7654:3210:FEDC:BA98:7654:3210]:80/index.html"),
        ("http://[::3210]:80/hi", "http://[::3210]:80/hi"),
        ("http://[3210:123::]:80/bye#", "http://[3210:123::]:80/bye#"),
        ("http://127.0.0.1/foo ", "http://127.0.0.1/foo"),
        ("www.ma-il.lv/$_", "www.ma-il.lv/$_"),
        ("http://url.com", "http://url.com"),
        ("http://funny.url.com", "http://funny.url.com"),
        ("http://url.com.", "http://url.com"),
        ("http://url.com./", "http://url.com./"),
        ("http://url.com.:123", "http://url.com.:123"),
        ("(http://url.com)", "http://url.com"),
        ("http://foo.com/blah_blah_(wikipedia)_(again))", "http://foo.com/blah_blah_(wikipedia)_(again)"),
        ("https://hp--community.force.com/", "https://hp--community.force.com/"),
        ("http://xn-d1abbgf6aiiy.xnpai/,", "http://xn-d1abbgf6aiiy.xnpai/"),
        ("http://xn--d1abbgf6aiiy.xn--p1ai/,", "http://xn--d1abbgf6aiiy.xn--p1ai/"),
        ("https://xn----8sbfxoeboc6b7i.xn--p1ai/", "https://xn----8sbfxoeboc6b7i.xn--p1ai/"),
        ("  https://en.wikipedia.org/wiki/Bap_(food)\x01", "https://en.wikipedia.org/wiki/Bap_(food)"),
        ('"https://en.wikipedia.org/wiki/Bap_(food)"', "https://en.wikipedia.org/wiki/Bap_(food)"),
        (" http://server.com/www.server.com ", "http://server.com/www.server.com"),
        ("http://badutf8pcokay.com/%FF?%FE#%FF", "http://badutf8pcokay.com/%FF?%FE#%FF"),
        ("http://www.abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.com/", "http://www.abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.com/"),
        ("http://abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcde.abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.com", "http://abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcde.abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.com"),

        ("0HTTP://ПРЕЗИДЕНТ.РФ'", "HTTP://ПРЕЗИДЕНТ.РФ", "HTTP://xn--d1abbgf6aiiy.xn--p1ai"),
        ("https://моя-молитва.рф/", "https://моя-молитва.рф/", "https://xn----8sbfxoeboc6b7i.xn--p1ai/"),
        ("\x03www.猫.jp", "www.猫.jp", "www.xn--z7x.jp"),
        ("(https://ru.wikipedia.org/wiki/Мыло_(значения))", "https://ru.wikipedia.org/wiki/Мыло_(значения)", "https://ru.wikipedia.org/wiki/%D0%9C%D1%8B%D0%BB%D0%BE_(%D0%B7%D0%BD%D0%B0%D1%87%D0%B5%D0%BD%D0%B8%D1%8F)"),
        ("http://➡.ws/䨹", "http://➡.ws/䨹", "http://xn--hgi.ws/%E4%A8%B9"),
        ("http://➡.ws/♥?♥#♥'", "http://➡.ws/♥?♥#♥", "http://xn--hgi.ws/%E2%99%A5?%E2%99%A5#%E2%99%A5"),
        ("http://➡.ws/♥/pa%2Fth;par%2Fams?que%2Fry=a&b=c", "http://➡.ws/♥/pa%2Fth;par%2Fams?que%2Fry=a&b=c", "http://xn--hgi.ws/%E2%99%A5/pa%2Fth;par%2Fams?que/ry=a&b=c"),
        ("http://website.com/path/is%2fslash/!$&'()*+,;=:@/path?query=!$&'()*+,;=:@?query#fragment!$&'()*+,;=:@#fragment", "http://website.com/path/is%2fslash/!$&'()*+,;=:@/path?query=!$&'()*+,;=:@?query#fragment!$&'()*+,;=:@#fragment", "http://website.com/path/is%2Fslash/!$&'()*+,;=:@/path?query=!$&'()*+,;=:@?query#fragment!$&'()*+,;=:@%23fragment")
    )

    print("\ntesting urls…")
    errors = False
    for test in tests:
        if not isinstance(test, tuple):
            string, exact, safe = test, None, None
        elif len(test) == 2:
            string, exact, safe = test[0], test[1], test[1]
        else:
            string, exact, safe = test
        result = list(find_urls(string))
        e, s = (result[1].exact, result[1].safe) if len(result) == 3 else (None, None)
        if e != exact or s != safe:
            print("FAIL %s: %s → %s; %s → %s" % (string, exact, e, safe, s))
            errors = True
    print("ERRORS FOUND" if errors else "no errors found")

    print("\ntesting speed…")
    iterations, repeat = 1000, 10
    import timeit
    urls = [test[0] for test in tests if isinstance(test, tuple) and "abcdefghijklm" not in test[0]]
    string = " lorem ipsum dolor sit amet ".join(urls)
    time = min(timeit.repeat(stmt="list(find_urls(text))",
                             setup="from %s import find_urls; text = u'''%s''' " % (__name__, string),
                             repeat=repeat,
                             number=iterations))
    print("%d loops, best of %d: %.5fms per loop (%d urls in %d characters)" %
          (iterations, repeat, time / iterations * 1000, len(urls), len(string)))

try:
    import weechat                                  # type: ignore
except ImportError:
    run_tests()
else:
    install()



# -*- coding: utf-8 -*-
#
# Script Name: Zerotab.py
# Script Author: Lucian Adamson <lucian.adamson@yahoo.com>
# Script License: GPL
# Alternate Contact: Freenode IRC nick i686
#
# 2013-01-27, Nils Görs <weechatter@arcor.de>:
#     version 1.5: make script compatible with Python 3.x
# 2011-09-20, Nils Görs <weechatter@arcor.de>:
#     version 1.4: fixed: latest nick from join/part messages were used.
# 2010-12-04, Sebastien Helleu <flashcode@flashtux.org>:
#     version 1.3: use tag "nick_xxx" (WeeChat >= 0.3.4 only)
# 2010-08-03, Sebastien Helleu <flashcode@flashtux.org>:
#     version 1.2: fix bug with nick prefixes (op/halfop/..)
# 2010-08-03, Sebastien Helleu <flashcode@flashtux.org>:
#     version 1.1: fix bug with self nick


SCRIPT_NAME='zerotab'
SCRIPT_AUTHOR='Lucian Adamson <lucian.adamson@yahoo.com>'
SCRIPT_VERSION='1.5'
SCRIPT_LICENSE='GPL'
SCRIPT_DESC='Will tab complete the last nick in channel without typing anything first. This is good for rapid conversations.'

import_ok=True

try:
    import weechat, re
except ImportError:
    print ('This script must be run under WeeChat')
    print ('You can obtain a copy of WeeChat, for free, at http://www.weechat.org')
    import_ok=False

latest_speaker={}
weechat_version=0

def my_completer(data, buffer, command):
    global latest_speaker
    str_input = weechat.buffer_get_string(weechat.current_buffer(), "input")
    if command == "/input complete_next" and str_input == '':
        nick = latest_speaker.get(buffer, "")
        if nick != "":
            weechat.command(buffer, "/input insert " + nick)
    return weechat.WEECHAT_RC_OK

def hook_print_cb(data, buffer, date, tags, displayed, highlight, prefix, message):
    global latest_speaker
    alltags = tags.split(',')
    if 'notify_message' in alltags:
        nick = None
        if int(weechat_version) >= 0x00030400:
            # in version >= 0.3.4, there is a tag "nick_xxx" for each message
            for tag in alltags:
                if tag.startswith('nick_'):
                    nick = tag[5:]
                    break
        else:
            # in older versions, no tag, so extract nick from printed message
            # this is working, except for irc actions (/me ...)
            nick = prefix
            if re.match('^[@%+~*&!-]', nick):
                nick = nick[1:]
        if nick:
            local_nick = weechat.buffer_get_string(buffer, "localvar_nick")
            if nick != local_nick:
                latest_speaker[buffer] = nick
    return weechat.WEECHAT_RC_OK

if __name__ == "__main__" and import_ok:
    if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, "", ""):
        weechat_version = weechat.info_get("version_number", "") or 0
        weechat.hook_print("", "", "", 1, "hook_print_cb", "")
        weechat.hook_command_run('/input complete*', 'my_completer', '')



# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 by nils_2 <weechatter@arcor.de>
#
# last written: provide item to keep track of last buffer you wrote something
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# 2019-06-02: nils_2, (freenode.#weechat)
#       0.1 : initial release

try:
    import weechat,re

except Exception:
    print('This script must be run under WeeChat.')
    print('Get WeeChat now at: https://www.weechat.org/')
    quit()

SCRIPT_NAME     = 'last_written'
SCRIPT_AUTHOR   = 'nils_2 <weechatter@arcor.de>'
SCRIPT_VERSION  = '0.1'
SCRIPT_LICENSE  = 'GPL'
SCRIPT_DESC     = 'keep track of last buffer you wrote something'

item_last_written = 'last_written'
item_last_sent = 'last_sent'
last_written = ''
last_sent = ''
# ==============================[ callback ]=============================
# signal_data = buffer_ptr
def input_text_changed_cb(data, signal, signal_data):
    global last_written
    buffer_name = weechat.buffer_get_string(signal_data, 'name')
    if last_written == buffer_name:
        return weechat.WEECHAT_RC_OK
    last_written = buffer_name
    update_item_cb(data,signal,signal_data,'last_written')
    return weechat.WEECHAT_RC_OK

# signal = buffer_ptr
def input_return_cb(data, signal, signal_data):
    global last_sent
    buffer_name = weechat.buffer_get_string(signal, 'name')
    if last_sent == buffer_name:
        return weechat.WEECHAT_RC_OK
    last_sent = buffer_name
    update_item_cb(data,signal,signal_data,'last_sent')
    return weechat.WEECHAT_RC_OK
# ================================[ item ]===============================
def bar_item_last_written_cb(data, item, window):
    global last_written
    # check for root input bar!
    if not window:
        window = weechat.current_window()

    # get current buffer (for example for split windows!)
    ptr_buffer = weechat.window_get_pointer(window,'buffer')
    if ptr_buffer == '':
        return ''
    return last_written

def bar_item_last_sent_cb(data, item, window):
    global last_written
    # check for root input bar!
    if not window:
        window = weechat.current_window()

    # get current buffer (for example for split windows!)
    ptr_buffer = weechat.window_get_pointer(window,'buffer')
    if ptr_buffer == '':
        return ''
    return last_sent

def update_item_cb(data, signal, signal_data,item_name):
    weechat.bar_item_update(item_name)
    return weechat.WEECHAT_RC_OK

def last_written_info_cb(data, info_name, arguments):
    global last_written
    return last_written

def last_sent_info_cb(data, info_name, arguments):
    global last_sent
    return last_sent

# ================================[ main ]===============================
if __name__ == '__main__':
    global version
    if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, '', ''):
        version = weechat.info_get('version_number', '') or 0

        weechat.hook_command(SCRIPT_NAME, SCRIPT_DESC, 'last_written||last_sent',
                            'script provide two items:\n'
                            '"last_written", will print name of buffer you typed text last time\n'
                            '"last_sent", will print name of buffer you sent text last time\n\n'
                            'You can use both items with /eval or in /key using variable "${info:last_written}" and "${info:last_sent}"\n\n'
                            'Example:\n'
                            'bind key to jump to last buffer you sent text\n'
                            ' /key bind meta-# /eval /buffer ${info:last_sent}\n'
                            'creates an item for text_item.py script (item name ""ti_last_written"\n'
                            ' /set plugins.var.python.text_item.ti_last_written "all|input_text_changed ${info:last_written} ${info:last_sent}"'
                            '',
                            '','','')

        weechat.bar_item_new(item_last_written, 'bar_item_last_written_cb','')
        weechat.bar_item_new(item_last_sent, 'bar_item_last_sent_cb','')
        weechat.hook_info('last_written',
                      'Return name of last buffer text was written',
                      '',
                      'last_written_info_cb', '')
        weechat.hook_info('last_sent',
                      'Return name of last buffer text was sent',
                      '',
                      'last_sent_info_cb', '')

        weechat.hook_command_run('/input return', 'input_return_cb', '')
        weechat.hook_signal ('input_text_changed', 'input_text_changed_cb', '')



# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2019 by nils_2 <weechatter@arcor.de>
#
# Display size of current logfile in item-bar
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# 2019-07-12: nils_2 (freenode.#weechat)
#       0.4 : option "display" is evaluated
#           : use hook_process("wc") to not stall weechat anymore
#           : new function: refresh only on buffer/window switch
#           : make script compatible with Python 3.x
# 2013-01-07: nils_2 (freenode.#weechat)
#       0.3 : missing logfile caused a crash (thanks swimmer)
#           : add support of more than one window
#           : two new options "log_disabled" and "file_not_found"
# 2012-11-22: nils_2 (freenode.#weechat)
#       0.2 : bug on first startup removed (thanks swimmer)
# 2012-01-14: nils_2 (freenode.#weechat)
#       0.1 : initial release
#
# How to use:
# add item "logsize" to option "weechat.bar.status.items"
#
# Development is currently hosted at
# https://github.com/weechatter/weechat-scripts

from __future__ import print_function
from builtins import str

try:
    import weechat, re

except Exception:
    print("This script must be run under WeeChat.")
    print("Get WeeChat now at: https://weechat.org")
    quit()

SCRIPT_NAME     = "logsize"
SCRIPT_AUTHOR   = "nils_2 <weechatter@arcor.de>"
SCRIPT_VERSION  = "0.4"
SCRIPT_LICENSE  = "GPL"
SCRIPT_DESC     = "display size of current logfile in item-bar"
OPTIONS         = { "refresh"       : ("0","refresh timer (in seconds). 0 = refresh only on buffer or window switch. this is the default setting"),
                    "size"          : ("KB","display length in KB/MB/GB/TB. Leave option empty for byte"),
                    "display"       : ("%L","possible item: %W = words, %L = lines or %F = file length. (content is evaluated, e.g. you can use colors with format \"${color:xxx}\", see /help eval)"),
                    "log_disabled"  : ("","displays a text in item, when logger is disabled for buffer"),
                    "file_not_found": ("","displays a text in item, when logfile wasn't found"),
                    }

hooks           = { "timer": "", "bar_item": "" }
hook_process_out = []
output          = ""

# regexp to match ${color} tags
regex_color=re.compile('\$\{([^\{\}]+)\}')

# regexp to match ${optional string} tags
regex_optional_tags=re.compile('%\{[^\{\}]+\}')

# ================================[ size ]===============================
def sizecheck(filesize):
    filesize = int(filesize)
    if OPTIONS["size"].lower() == "kb":
        filesize = "%.2f" % (filesize / 1024)
        size = "K"
    elif OPTIONS["size"].lower() == "mb":
        filesize = "%.2f" % (filesize / 1024 / 1024)
        size = "M"
    elif OPTIONS["size"].lower() == "gb":
        filesize = "%.2f" % (filesize / 1024 / 1024 / 1024)
        size = "G"
    elif OPTIONS["size"].lower() == "tb":
        filesize = "%.2f" % (filesize / 1024 / 1024 / 1024 / 1024)
        size = "T"
    else:
        filesize = "%.0f" % filesize
        size = "b"
    return "%s%s" % (filesize,size)

# ================================[ weechat item ]===============================
def show_item (data, item, window):
    global output
    return output

def get_file_information(ptr_buffer):
    global hook_process_out, output

    (logfile,log_enabled) = get_logfile(ptr_buffer)
    if not log_enabled:
        output = OPTIONS["log_disabled"]
        return

    output = ''
    if logfile != '':
        # newline / word / bytes / filename
        weechat.hook_process("wc %s" % logfile, 50000, "my_hook_process_cb", "")
        if hook_process_out:
            lines = hook_process_out[0]
            words = hook_process_out[1]
            flength = sizecheck(hook_process_out[2])

            tags = {'%L': str(lines),
                    '%W': str(words),
                    '%F': str(flength)}

            output = substitute_colors(OPTIONS['display'])
            # replace mandatory tags
            for tag in list(tags.keys()):
                #    for tag in tags.keys():
                output = output.replace(tag, tags[tag])
    weechat.bar_item_update(SCRIPT_NAME)
    return

def substitute_colors(text):
    if int(version) >= 0x00040200:
        return weechat.string_eval_expression(text,{},{},{})
    # substitute colors in output
    return re.sub(regex_color, lambda match: weechat.color(match.group(1)), text)

def get_logfile(ptr_buffer):
    log_filename = ""
    log_enabled = 0
    infolist = weechat.infolist_get('logger_buffer','','')
    while weechat.infolist_next(infolist):
        bpointer = weechat.infolist_pointer(infolist, 'buffer')
        if ptr_buffer == bpointer:
            log_filename = weechat.infolist_string(infolist, 'log_filename')
            log_enabled = weechat.infolist_integer(infolist, 'log_enabled')
            log_level = weechat.infolist_integer(infolist, 'log_level')
    weechat.infolist_free(infolist)                  # free infolist()

    return (log_filename,log_enabled)

def item_update(data, remaining_calls):
    global hooks
    weechat.bar_item_update(SCRIPT_NAME)
    return weechat.WEECHAT_RC_OK

# ================================[ hook process]===============================
def my_hook_process_cb(data, command, return_code, out, err):
    global hook_process_out
    if return_code == weechat.WEECHAT_HOOK_PROCESS_ERROR:
        weechat.prnt("", "Error with command '%s'" % command)
        return weechat.WEECHAT_RC_OK
#    if return_code >= 0:
#        weechat.prnt("", "return_code = %d" % return_code)
    if out != "":
        hook_process_out = out.split()
    if err != "":
        weechat.prnt("", "stderr: %s" % err)
    return weechat.WEECHAT_RC_OK

# ================================[ weechat hook ]===============================
def window_switch_cb(data, signal, signal_data):
    window = signal_data
    window = weechat.current_window()
    ptr_buffer = weechat.window_get_pointer(window,"buffer")
    get_file_information(ptr_buffer)
    return weechat.WEECHAT_RC_OK

def buffer_switch_cb(data, signal, signal_data):
    ptr_buffer = signal_data
    window = weechat.current_window()
    ptr_buffer = weechat.window_get_pointer(window,'buffer')
    if ptr_buffer == '':
        return ''
    get_file_information(ptr_buffer)
    return weechat.WEECHAT_RC_OK

def hook_timer_refresh_item_cb(data, remaining_calls):
    weechat.bar_item_update(SCRIPT_NAME)
    return weechat.WEECHAT_RC_OK

def unhook_timer():
    global hooks
    if hooks["timer"] != "":
        weechat.bar_item_remove(hooks["bar_item"])
        weechat.unhook(hooks["timer"])
        hooks["timer"] = ""
        hooks["bar_item"]

def hook_timer():
    global hooks
    hooks["timer"] = weechat.hook_timer(int(OPTIONS["refresh"]) * 1000, 0, 0, 'item_update', '')
    hooks["bar_item"] = weechat.bar_item_new(SCRIPT_NAME, 'show_item','')

    if hooks["timer"] == 0:
        weechat.prnt('',"%s: can't enable %s, hook failed" % (weechat.prefix("error"), SCRIPT_NAME))
        weechat.bar_item_remove(hooks["bar_item"])
        hooks["bar_item"] = ""
        return 0
    weechat.bar_item_update(SCRIPT_NAME)
    return 1

# ================================[ weechat options and description ]===============================
def toggle_refresh(pointer, name, value):
    global hooks
    option = name[len('plugins.var.python.' + SCRIPT_NAME + '.'):]        # get optionname
    OPTIONS[option] = value                                               # save new value

    if option == 'refresh':                                               # option "refresh" changed by user?
        if hooks["timer"] != "":                                          # timer currently running?
            if OPTIONS['refresh'] != "0":                                 # new user setting not zero?
                unhook_timer()
                hook_timer()
            else:
                unhook_timer()                                            # user switched timer off
        elif hooks["timer"] == "":                                        # hook is empty
            if OPTIONS['refresh'] != "0":                                 # option is not zero!
                hook_timer()                                              # install hook
    weechat.bar_item_update(SCRIPT_NAME)
    return weechat.WEECHAT_RC_OK

def init_options():
    global OPTIONS

    for option, value in list(OPTIONS.items()):
        weechat.config_set_desc_plugin(option, '%s (default: "%s")' % (value[1], value[0]))
        if not weechat.config_is_set_plugin(option):
            weechat.config_set_plugin(option, value[0])
            OPTIONS[option] = value[0]
        else:
            OPTIONS[option] = weechat.config_get_plugin(option)
# ================================[ main ]===============================
if __name__ == "__main__":
  if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, '', ''):
      version = weechat.info_get("version_number", "") or 0

      init_options()

      if OPTIONS["refresh"] != "0":
          hook_timer()
      else:
          weechat.hook_signal("buffer_switch","buffer_switch_cb","")
          weechat.hook_signal("window_switch","window_switch_cb","")
          weechat.bar_item_new(SCRIPT_NAME, 'show_item','')
          weechat.bar_item_update(SCRIPT_NAME)
      weechat.hook_config( 'plugins.var.python.' + SCRIPT_NAME + '.*', 'toggle_refresh', '' )



# Copyright (c) 2015 by timss <timsateroy@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

SCRIPT_NAME = 'lastfm2'
SCRIPT_AUTHOR = "timss <timsateroy@gmail.com>"
SCRIPT_VERSION = '0.1'
SCRIPT_LICENSE = 'GPL3'
SCRIPT_DESC = "Sends latest played track for a Last.fm user to the current buffer"

SCRIPT_COMMAND = 'lastfm'
SCRIPT_HELP = \
"""Sends latest played track for a Last.fm user to the current buffer.

    /lastfm

By default, the script will use the username set in {SCRIPT_NAME} configuration:

    /set plugins.var.python.{SCRIPT_NAME}.user yourusername

In addition, an username may be specified as an argument:

    /lastfm anotherusername

The command which output will be sent to the buffer may be customized as well:

    /set plugins.var.python.{SCRIPT_NAME}.command I'm listening to {{track}}

Finally, the command when specifying another username can also be set:

    /set plugins.var.python.{SCRIPT_NAME}.command_arg {{user}} is litening to {{track}}

Inspiration and credit:
    - lastfm.py, Adam Saponara <saponara TA gmail TOD com>
    - lastfmnp.py, i7c <i7c AT posteo PERIOD de>
    - lastfmapi.py, Christophe De Troyer <christophe@call-cc.be>

""".format(SCRIPT_NAME=SCRIPT_NAME)

try:
    import weechat
    import_ok = True
except ImportError:
    print("This script must be run under WeeChat.")
    print("Get WeeChat now at: http://www.weechat.org/")
    import_ok = False

import json

def init_config():
    """Set plugin options to defaults if not already done"""
    config = {
        'user': '',
        'command': '/me is listening to {track}',
        'command_arg': '{user} is listening to {track}',
        'api_key': 'ae51c9df97d4e90c35ffd302e987efd2',
        'api_url': 'https://ws.audioscrobbler.com/2.0/?method=user.getRecentTracks&user={user}&limit=1&api_key={api_key}&format=json',
        'timeout': '10000'
    }

    for option, default in config.items():
        if not weechat.config_is_set_plugin(option):
            weechat.config_set_plugin(option, default)

def get_recent_track(data, command, rc, out, err):
    """Get last track played (artist - name)"""
    if rc == weechat.WEECHAT_HOOK_PROCESS_ERROR:
        weechat.prnt('', "Error with command '{}'".format(command))
    elif rc > 0:
        weechat.prnt('', "rc = {}".format(rc))

    try:
        data = json.loads(out)

        if data.has_key('error'):
            weechat.prnt('', "Last.fm API error: '{}'".format(data['message']))
        else:
            artist = data['recenttracks']['track'][0]['artist']['#text']
            name = data['recenttracks']['track'][0]['name']
            track = "{} - {}".format(artist, name)
            user = data['recenttracks']['@attr']['user'].lower()

            # print username or not, depending on config/arg
            if user == weechat.config_get_plugin('user').lower():
                cmd = weechat.config_get_plugin('command')
            else:
                cmd = weechat.config_get_plugin('command_arg')

            # format isn't picky, ignores {user} if not present
            cmd = cmd.format(user=user, track=track)

            weechat.command(weechat.current_buffer(), cmd)
    except IndexError, KeyError:
        weechat.prnt('', "Error parsing Last.fm data")

    return weechat.WEECHAT_RC_OK

def lastfm_cmd(data, buffer, args):
    """Print last track played"""
    api_key = weechat.config_get_plugin('api_key')
    api_url = weechat.config_get_plugin('api_url')
    timeout = weechat.config_get_plugin('timeout')

    # use user in argument, or in config
    if args:
        user = args
    else:
        user = weechat.config_get_plugin('user')

    url = 'url:' + api_url.format(user=user.lower(), api_key=api_key)
    weechat.hook_process(url, int(timeout), 'get_recent_track', '')

    return weechat.WEECHAT_RC_OK

if __name__ == '__main__' and import_ok:
    if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, '', ''):
        init_config()
        weechat.hook_command(SCRIPT_COMMAND, SCRIPT_HELP, '', '', '', 'lastfm_cmd', '')




#       quodnp.py
#
#       Copyright 2009 Brandon Hartshorn <sharntehnub AT gmail DOT com>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.

import weechat, os, sys, re
from stat import *

SCRIPT_NAME    = "quodnp"
SCRIPT_AUTHOR  = "Sharn"
SCRIPT_VERSION = "0.5"
SCRIPT_LICENSE = "GPL2"
SCRIPT_DESC    = "Full control of Quodlibet from Weechat"

SCRIPT_COMMAND = "quodnp"

if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE,
                    SCRIPT_DESC, "", ""):
    weechat.hook_command(
        SCRIPT_COMMAND,
        "Control of Quodlibet in Weechat",
        "np | next | prev | play-pause",
        "            np: print song now playing to current buffer\n"
        "next/prev/play: control quodlibet - next song/previous song/play-pause"
        " respectively\n"
        "For np_format configuration, you can use anything avaliable"
        "from \"cat ~/.quodlibet/current\" - just ignore the ~#s\n"
        "and add a \"$\" - for example, to print the artist put $artist, or $album for album.",
        "np|next|prev|play-pause",
        "command_handle", ""
        )

# default options
settings = {
    "autonp"                    : "off",
    "np_format"                 : "np: $artist - $title",
    "debug"                     : "off",
}

for option, default_value in settings.items():
    if weechat.config_get_plugin(option) == "":
        weechat.config_set_plugin(option, default_value)

def quodlibet_nowplaying(buffer):
    values = {}

    current_file = os.path.expanduser("~/.quodlibet/current")
    if os.path.isfile(current_file):
        open_file = open(current_file, "r")
        for line in open_file:
            key, val = line.lstrip("~#").strip().split("=", 1)
            if key == "bitrate":
                val = val[:3] + "Kbps"
            elif key == "length":
                val = val
            values.update({
            key : val,
            })
        weechat.command(weechat.current_buffer(), (format_output(weechat.config_get_plugin("np_format"), values)))
        open_file.close()
    else:
        weechat.prnt("", "Error opening " + current_file + ". Are you sure Quodlibet is running?")

    ouput = format_output(weechat.config_get_plugin("np_format"), values)

def quodlibet_control(action):
    control_file = os.path.expanduser("~/.quodlibet/control")
    error = "Error opening " + control_file + ". Are you sure Quodlibet is running?"
    try:
        mode = os.stat(control_file)[ST_MODE]
        if S_ISFIFO(mode):
            open_file = open(control_file, "w")
            open_file.write(action)
            open_file.close()
        else:
            weechat.prnt("", error)
    except:
        weechat.prnt("", error)

def format_output(format, values):
  out = ""
  n = 0
  for match in re.finditer(findvar, format):
    if match is None: continue
    else:
      l, r = match.span()
      nam = match.group(1)
      out += format[n:l+1] + values.get(nam, "").strip()
      n = r
  return out + format[n:]

findvar = re.compile(r'[^\\]\$([a-z_]+)(\b|[^a-z_])')

def command_handle(data, buffer, args):
    largs = args.split(" ")
    if len(largs) > 1:
        weechat.prnt("", "This script can only use 1 argument at a time, see /help " + SCRIPT_COMMAND + " if you need help")
    elif largs[0] in ("next", "prev", "play-pause"):
        quodlibet_control(largs[0])
    elif largs[0] == "np":
        quodlibet_nowplaying(buffer)
    else:
        if weechat.config_get_plugin("autonp") == "on":
            quodlibet_nowplaying(buffer)
        else:
            weechat.prnt("", "No action specified, see /help " + SCRIPT_COMMAND + " if you need help")
    return weechat.WEECHAT_RC_OK



# -*- coding: utf-8 -*-
#
# The MIT License (MIT)
#
# Copyright (c) 2015 - 2016 Jos Ahrens <zarthus@lovebytes.me>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# ---
#
# Whenever a mode with a hostmask argument is set (like a ban, or a quiet), this
# script compares the mask to anyone in the current channel to see if it matches
# anyone, and lists them accordingly.
#
# This script comes with a command (/maskmatch <hostname>) to perform a match
# before you perform a ban or quiet on anyone, and as well has partial extban support
# in the form of recognizing account names with $a:accountname and $~a
#
# Script options:
#   whitelist, blacklist (default: "", options: see description)
#     The whitelist supports multiple types of arguments. Primarily a server,
#     channel, or combination of the two.
#
#     examples:
#       freenode will match all channels on freenode
#       #freenode will match the channel #freenode on all networks
#       freenode.#freenode will match the channel #freenode on freenode
#
#     Matching objects on the whitelist will only match on these targets.
#     Matching objects on the blacklist will never match on these targets.
#     When both options are set, the whitelist takes precedence over the blacklist.
#     When neither option is set, everything matches.
#
#   disabled (default: false, options: true, false)
#     Relatively straightforward, if you want to disable the listing of users
#     when someone sets a ban, and are just interested in the /modematch command,
#     set this to true.
#
#   ignore_masks (default: *!*@*, options: comma separated string of hostmasks)
#     Some masks (in particular *!*@*) match a large array of users. Hostmasks in
#     this list should never match.
#
#   limit (default: 4, options: integer)
#     The limit indicates how many matches to print before aborting. This is to
#     prevent flood on larger matches. The limit cannot be turned off (it will always
#     be at least 1), but can be set to arbitrarily high numbers.
#
#     When the limit is reached, you will see "... and x more matches".
#
#   match_set, match_unset (default: true, options: true, false)
#     Configure if you want to match only when a mode is set or unset.
#
#     The script will not show matches when both options are set to false.
#
#   matching_modes (default: beIq, options: list of mode characters)
#     List of modes that have a hostmask parameter. In version 1.0, maskmatch does
#     not read the server's isupport to see which modes have parameters to them.
#
#     Because it does not do this, problems may occur if you are on a network that, for example,
#     supports +q as "channel owner" as opposed to quiet. You could either mark the network as
#     blacklisted, or use this option to negate all q matches. In addition to that, you may simply
#     not wish to see matches for invite exceptions or quiets.
#
#   prefix (default: [_script_name_], options: see description)
#     The prefix is the name the script assumes in front of the printing message.
#     This option can be any type of string, inherit from network_prefix, or more.
#     Multiple special options can be in place.
#
#     In addition to this, if the prefix is surrounded in [] brackets, the prefix_color
#     will be assumed upon the brackets, rather than the prefix itself.
#
#     Special options:
#       _script_name_    - Gets converted into SCRIPT_NAME, which would be maskmatch
#       _prefix_network_ - The option you have set in weechat.look.prefix_network
#       _setter_         - The person whom set the mode
#       _target_         - The person whom the mode was set on. This is their nick, not the hostmask
#
#     The _setter_ and _target_ options do not honor prefix_color, and will always have their own
#     weechat color.
#
#   prefix_color (default: green, options: any weechat supported color)
#     The prefix color is the color the prefix (or the brackets surrounding the prefix)
#     will have. This can be any weechat supported color.
#
#   print_as_list (default: false, options: true, false)
#     Instead of printing a newline per message, print it as a large list instead.
#     This formatting is more compact, and will consume less space. But may as direct
#     result be less readable for larger limits.
#
#   sorting (default: alpha_ignore_case, options: none, alpha, alpha_ignore_case)
#     Sort matches by their name - either alphabetically, alphabetically with case insensitivity,
#     or none at all.
#
#   verbose (default: false, options: true, false)
#     When verbose is on, you will be informed when there aren't any matches found.
#
# This script functions on WeeChat 1.3 and above, supporting both Python 2 and Python 3.
# It probably runs on earlier versions of WeeChat, but has not been tested.
#
# History:
#  version 1.0 - 2015-12-19
#    Script creation
#  version 1.1 - 2015-12-19
#    cmd_maskmatch: Validate if existing buffer is a channel
#    match_against_nicklist:
#      Fix wrong comparison.
#      Optimize iterating through the nicklist by not constantly reperforming some
#        calculations, and instead storing them into boolean values.
#    Code formatting fixes
#  version 1.2 - 2015-12-31
#    bug fix: Add missing parenthesis to ensure proper .format() calls.
#  version 1.3 - 2016-09-21
#    consistency: honour weechat.look.prefix_same_nick in print_as_list(), also now
#      use the name in both messages of print_as_list(). This change is slightly
#      backwards-incompatible. You will need to edit either
#      weechat.look.prefix_same_nick with a space (" ") to get the old behaviour,
#      or set the script option prefix to a space (also not ideal.)
#
#      Feel free to provide feedback on this change if you liked the old behaviour
#      and don't find the alternate solutions acceptable. Considering the size of
#      this change you could easily revert back to 1.2. Future releases will provide
#      a solution if the old behaviour was liked.
#  version 1.4 - 2016-10-11
#    bug fix: works around a problem where modes with arguments are not being filtered out,
#      and instead maskmatch starts matching against the wrong argument.
#      This is a bit of a cheesy hack. Anyone is welcome to improve the logic on this.
#      relevant: https://github.com/Zarthus/weechat-scripts/issues/3

try:
    import weechat as w
    import_ok = True
except ImportError:
    print("This script must be run under WeeChat")
    print("Get WeeChat now at: https://weechat.org/")
    import_ok = False


SCRIPT_NAME = "maskmatch"
SCRIPT_AUTHOR = "Zarthus <zarthus@lovebytes.me>"
SCRIPT_VERSION = "1.4"
SCRIPT_LICENSE = "MIT"
SCRIPT_DESC = "Display who got banned (quieted, excepted, etc.) when a mode with a hostmask argument is set"
SCRIPT_COMMAND = "maskmatch"


def cmd_maskmatch(data, buffer, mask):
    """Trigger for /maskmatch <hostmask>"""

    is_channel = False
    try:
        server, channel = w.buffer_get_string(buffer, "name").split(".", 1)
        is_channel = True
    except ValueError:
        is_channel = False
        w.prnt(buffer, "error: Active buffer does not appear to be a channel.")

    if not is_channel:
        return w.WEECHAT_RC_OK

    matches = match_against_nicklist(server, channel, mask)

    print_matches(buffer, matches, {"setter": "maskmatch", "mode": "special", "set": True, "mask": mask})

    return w.WEECHAT_RC_OK

def on_channel_mode(data, signal, signal_data):
    """Whenever a channel mode is set."""

    if w.config_get_plugin("disabled") in ["true", "yes"]:
        return w.WEECHAT_RC_OK

    parsed = w.info_get_hashtable("irc_message_parse", {"message": signal_data})

    server = signal.split(",")[0]
    channel = parsed["channel"]

    if not should_match(server, channel):
        return w.WEECHAT_RC_OK

    chars = w.config_get_plugin("matching_modes")
    modes_set = parsed["text"].split(" ")[0]
    found = False

    for c in chars:
        if c in modes_set:
            found = True
            break

    if not found:
        return w.WEECHAT_RC_OK

    modes = parse_modes(parsed["text"])

    for mode in modes:
        mode["setter"] = parsed["nick"]
        match_mode(server, channel, mode)

    return w.WEECHAT_RC_OK


def match_mode(server, channel, data):
    """When the matching mode should be compared against the userlist."""

    if is_mask_ignored(data["mask"]):
        return False

    if w.config_get_plugin("match_unset") not in ["true", "yes"] and not data["set"]:
        return False

    if w.config_get_plugin("match_set") not in ["true", "yes"] and data["set"]:
        return False

    target = w.buffer_search("irc", "{}.{}".format(server, channel))

    matches = match_against_nicklist(server, channel, data["mask"])
    print_matches(target, matches, data)

    return True


def parse_modes(text):
    """Go through the mode string (+bbb arg1 arg2 arg3) and split them up, discarding unsupported modes"""

    modes, args = text.split(" ", 1)
    masks = args.split(" ")
    chars = w.config_get_plugin("matching_modes")

    toggle = None
    i = 0
    ret = []

    for c in modes:
        if c == "+" or c == "-":
            toggle = c == "+"
            continue

        if c not in chars:
            if c in ["I", "k", "e", "b", "q"] or not is_maskmatch_mask(masks[i]):  # TODO: look in isupport CHANMODES and PREFIX
                del masks[i]
            continue

        ret.append({"set": toggle, "mode": c, "mask": masks[i]})
        i += 1

    return ret


def match_list_item(server, channel, item):
    """Match a list item against a server.channel, channel, or server combination."""

    if "." in item:
        if "{}.{}".format(server, channel) == item:
            return True
    else:
        if channel == item or server == item:
            return True

    return False


def should_match(server, channel):
    """Validate if the server.channel combination is in any of the white or blacklist items."""

    whitelist = w.config_get_plugin("whitelist")
    blacklist = w.config_get_plugin("blacklist")

    if not whitelist and not blacklist:
        return True

    if whitelist:
        wl = whitelist.split(",")
        for item in wl:
            if match_list_item(server, channel, item):
                return True
        return False
    else:
        bl = blacklist.split(",")
        for item in bl:
            if match_list_item(server, channel, item):
                return False
        return True

def is_maskmatch_mask(mask):
    """Validate if a mask is a valid IRC mask we support"""

    # author comment: if we ever need regex for more than just
    # this, this statement could be replaced with a regex
    # that is more accurate. /\$~a|\$a:[^ ]+|[^! ]+![^@ ]+@/ or the likes.

    return "$~a" in mask or "$a:" in mask or ("!" in mask and "@" in mask)


def is_mask_ignored(mask):
    """Validate if banmask is in the ignored banmask list."""

    ignored = w.config_get_plugin("ignore_masks").split(",")

    for banmask in ignored:
        if mask == banmask:
            return True
    return False


def match_against_nicklist(server, channel, hostmask):
    """Compare the hostmask against all users in the channel"""

    infolist = w.infolist_get("irc_nick", "", "{},{}".format(server, channel))

    if "$a:" in hostmask or "$~a" in hostmask:
        field = "account"
        hostmask = hostmask.replace("$a:", "")
        hostfield = False
    else:
        field = "host"
        hostfield = True

    extban_unreg = hostmask == "$~a"
    matches = []

    while w.infolist_next(infolist):
        name = w.infolist_string(infolist, "name")

        if hostfield:
            host = name + "!" + w.infolist_string(infolist, field)
        else:
            host = w.infolist_string(infolist, field)

        if ((extban_unreg and host == "*") or
            (not extban_unreg and w.string_match(host, hostmask, 0))):
            matches.append(name)

    w.infolist_free(infolist)
    return matches


def print_matches(target, matches, data):
    """Print all matching masks to the target channel"""

    verbose = w.config_get_plugin("verbose") in ["true", "yes"]
    limit = int(w.config_get_plugin("limit"))

    if limit < 1:
        limit = 1

    total = len(matches)
    if total == 0:
        if verbose or data["mode"] == "special":
            w.prnt(target, "{}\tNo matches for {}".format(fmt_prefix(data).replace("_target_", ""), fmt_banmask(data["mask"])))
        return

    sorting = w.config_get_plugin("sorting")
    if sorting == "alpha":
       matches = sorted(matches)
    elif sorting == "alpha_ignore_case":
       matches = sorted(matches, key=str.lower)

    if w.config_get_plugin("print_as_list") in ["true", "yes"]:
        print_as_list(target, matches, data, limit, total)
    else:
        print_as_lines(target, matches, data, limit, total)

def print_as_list(target, matches, data, limit, total):
    """Prints the output as a comma-separated list of nicks."""

    col = w.color(w.info_get("irc_nick_color_name", data["setter"]))
    pf = fmt_prefix(data).replace("_target_", "")

    s = "{}\tThe following {} {}"
    if data["mode"] == "special":
        w.prnt(target, s.format(pf, "nick matches" if total == 1 else "nicks match", fmt_banmask(data["mask"])))
    else:
        w.prnt(target, (s + ", {} by {}{}{}").format(
           pf, "nick matches" if total == 1 else "nicks match",
           fmt_banmask(data["mask"]), fmt_mode_char(data["mode"]), col,
           data["setter"], w.color("reset")
        ))

    nicks = []
    remainder = len(matches) - limit
    i = 0
    for name in matches:
       nicks.append("{}{}{}".format(w.color(w.info_get("irc_nick_color_name", name)), name, w.color("reset")))
       i += 1

       if i >= limit:
           break

    if w.config_string(w.config_get("weechat.look.prefix_same_nick")):
        pf = (w.color(w.config_get_plugin("prefix_color")) +
          w.config_string(w.config_get("weechat.look.prefix_same_nick")) +
          w.color("reset"))

    printstr = "{}\t{}".format(pf, ", ".join(nicks))
    if remainder > 0:
        printstr += ", and {} more..".format(remainder)
    w.prnt(target, printstr)


def print_as_lines(target, matches, data, limit, total):
    """Prints the output as a line-separated list of nicks."""

    prefix = fmt_prefix(data)
    mstring = "{}{}".format(fmt_mode_char(data["mode"]), "" if data["set"] else " removal")
    mask = fmt_banmask(data["mask"])
    target_in_prefix = "_target_" in prefix
    i = 0

    for name in matches:
        if target_in_prefix:
            pf = prefix.replace("_target_", "{}{}{}".format(
                w.color(w.info_get("irc_nick_color_name", name)),
                name, w.color("reset")))
        else:
            pf = prefix

        if (total - (limit - i) == 1) or (i >= limit):
            left = total - i
            left -= 1 if target_in_prefix else 0

            w.prnt(target, "{}\tand {} more match{}..".format(pf, left, "es" if left != 1 else ""))
            break

        if target_in_prefix:
            w.prnt(target, "{}\tmatches {} {}".format(pf, mstring, mask))
        else:
            w.prnt(target, "{}\t{} {} matches {}".format(pf, mstring, mask, fmt_nick(name)))
        i += 1


def fmt_banmask(mask):
    """Formats a banmask"""

    green = w.color("green")
    reset = w.color("reset")

    return "{}[{}{}{}]{}".format(green, reset, mask, green, reset)


def fmt_mode_char(char):
    """Translate a mode character into a readable string"""

    chars = {
        "b": "ban",
        "e": "ban exception",
        "q": "quiet",
        "I": "invite exception",
        "special": "testing hostmask"
    }

    if char in chars:
        return chars[char]
    return "mode " + char


def fmt_nick(nick):
    """Format nick in colours for output colouring"""

    green = w.color("green")
    reset = w.color("reset")
    nick_col = w.color(w.info_get("irc_nick_color_name", nick))

    return "{}[{}{}{}]{}".format(green, nick_col, nick, green, reset)


def fmt_prefix(data):
    """Read and transform the prefix as per user settings"""

    fmt = w.config_get_plugin("prefix")

    if "_script_name_" in fmt:
        fmt = fmt.replace("_script_name_", SCRIPT_NAME)
    if "_setter_" in fmt:
        fmt = fmt.replace("_setter_", data["setter"])
    if "_prefix_network_" in fmt:
        fmt = fmt.replace("_prefix_network_", w.config_string(w.config_get("weechat.look.prefix_network")))
    col = w.config_get_plugin("prefix_color")


    pfcol = w.color(col)
    reset = w.color("reset")

    if w.string_match(fmt, "[*]", 0):
        fmt = fmt.replace("[", "{}[{}".format(pfcol, reset)).replace("]", "{}]{}".format(pfcol, reset))
    else:
        fmt = "{}{}{}".format(pfcol, fmt, reset)

    return fmt


if import_ok and w.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, "", ""):
    settings = {
        "blacklist": ["", "List of servers, channels, or server.channel combinations this should not be done for (whitelist takes precedence). Comma separated string"],
        "disabled": ["false", "Set this to true to entirely disable mode matching, and just allow use for the /maskmatch command. Boolean"],
        "ignore_masks": ["*!*@*", "Hostmasks that are too broad that should be ignored. Comma separated string"],
        "limit": ["4", "How many matches should be displayed at maximum? Number"],
        "match_set": ["true", "Match when a mode gets set. Boolean"],
        "match_unset": ["true", "Match when a mode gets unset. Boolean"],
        "matching_modes": ["bIqe", "List of mode characters to match on, every character represents a mode."],
        "prefix": ["_script_name_", "The name the script will have. Special options: _script_name_, _prefix_network_, _setter_, _target_"],
        "prefix_color": ["green", "The colour the prefix will have. If prefix is wrapped in brackets, the brackets will have this color. Any weechat-supported color"],
        "print_as_list": ["false", "Print as one large list, less information - but can fit more names at cost of readability. Boolean"],
        "sorting": ["alpha_ignore_case", "Sort names alphabetically or not. none, alpha, or alpha_ignore_case"],
        "verbose": ["false", "Print also if no matches are found. Boolean"],
        "whitelist": ["", "List of servers, channels, or server.channel combinations this should be done for. Comma separated string"]
   }

    for option, default_value in settings.items():
        if not w.config_is_set_plugin(option):
            w.config_set_plugin(option, default_value[0])
            w.config_set_desc_plugin(option, default_value[1])

    w.hook_command(
        SCRIPT_COMMAND,
        ("Compare a hostmask against the active channel to see how much people it would affect."
         "\n\nA fully qualified hostname should be used in most cases, as it might give unexpected results "
         "when this is not done."
         "\n\nMaskmatch compares against the users in the active channel."
         "\n\nYou can also use account extbans like so: $a:account, $~a."
         " Full extban support is not implemented."),
        "<hostmask>",
        ("The argument <hostmask> is the hostmask to compare against in the active channel, "
         "and should be a fully qualified hostname (*!*@* format), or an account extban."),
        "", "cmd_maskmatch", ""
    )

    w.hook_signal("*,irc_in_MODE", "on_channel_mode", "")




# coding: utf-8
#
# Copyright (c) 2012 by nesthib <nesthib@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# This script generates a file containing a formatted list of highlights to be
# used by an external program like conky.
#
# 2014-05-10, Sébastien Helleu <flashcode@flashtux.org>
#        0.2: change hook_print callback argument type of displayed/highlight
#             (WeeChat >= 1.0)
# 2012-03-07: nesthib <nesthib@gmail.com>
#        0.1: initial release

try:
    import weechat as w
except Exception:
    print("This script must be run under WeeChat.")
    print("Get WeeChat now at: http://www.weechat.org/")
    quit()

import re, os
from collections import Counter

name = "hl2file"
author = "nesthib <nesthib@gmail.com>"
version = "0.2"
license = "GPL"
description = "Generates a file with highlights for external programs like conky"
shutdown_function = "shutdown"
charset = ""

w.register(name, author, version, license, description, shutdown_function, charset)

settings = {
        'output_file': '%h/highlights.txt',
        'output_active_buffer': '%h/active_buffer.txt',
        #'ignore_chans': 'server',
        'output_format': '%{buffer}: %{nick} (%{message})',
        'summarize_by_buffer': 'off',
        'short_buffer_names': 'on',
        'clear_visited_buffers': 'on',
        }

for opt, val in settings.iteritems():
    if not w.config_is_set_plugin(opt):
        w.config_set_plugin(opt, val)

highlights = []
regex_tags = re.compile('%\{[^\{\}]+\}')

def bufname(buffer):
    buffer_name = w.buffer_get_string(buffer , 'name')
    if w.config_get_plugin('short_buffer_names') == 'on':
        buffer_name = buffer_name.partition('.')[2]
    return buffer_name

def write(my_file, content):
    if my_file == "":
        return
    my_file = os.path.expanduser(my_file.replace("%h", w.info_get("weechat_dir", "")))
    try:
        f = open(my_file, 'w')
    except IOError:
        w.prnt('', 'Error: %s is unable to open "%s", please select an appropriate location' % (name, my_file))
    else:
        f.writelines(content)
        f.close()

def my_print_cb(data, buffer, date, tags, displayed, highlight, prefix, message):
    global highlights
    if w.config_get_plugin('clear_visited_buffers') == 'on':
        if buffer == w.window_get_pointer(w.current_window(), 'buffer'):
            return w.WEECHAT_RC_OK
    if int(highlight):
        infos = {'buffer'   : buffer,
                 'date'     : date,
                 'tags'     : str.split(tags, ','),
                 'nick'   : prefix,
                 'message'  : message,
                 }
        highlights.append(infos)
        write_file()
    return w.WEECHAT_RC_OK

def write_file():
    global highlights
    filename = w.config_get_plugin('output_file')
    if w.config_get_plugin('summarize_by_buffer') == 'on':
        items = Counter([ bufname(item['buffer']) for item in highlights ])
        output_lines = [ '%s (%s)\n' % (item, items[item]) for item in items ]
    else:
        output_regex = w.config_get_plugin('output_format')
        tags = regex_tags.findall(output_regex)
        output_lines = []
        for infos in highlights:
            output_string = output_regex+'\n'
            for tag in tags:
                item = tag.lstrip('%{').rstrip('}')
                if item in infos.keys():
                    if item == 'buffer':
                        new_string = bufname(infos[item])
                    else:
                        new_string = infos[item]
                    output_string = output_string.replace(tag, new_string)
            output_lines.append(output_string)
    write(filename, output_lines)

def clear_file_cb(data, buffer, args):
    global highlights
    highlights = []
    write_file()
    return w.WEECHAT_RC_OK

def buffer_switch_cb(data, signal, signal_data):
    global highlights
    if w.config_get_plugin('clear_visited_buffers') == 'on':
        buffer = w.window_get_pointer(w.current_window(), 'buffer')
        write(w.config_get_plugin('output_active_buffer'), bufname(buffer))
        highlights = [ item for item in highlights if item['buffer'] != buffer ]
        write_file()
    return w.WEECHAT_RC_OK

def shutdown():
    for my_file in ['output_file', 'output_active_buffer']:
        filename = w.config_get_plugin(my_file).replace("%h", w.info_get("weechat_dir", ""))
        if os.path.exists(filename):
            os.remove(filename)
    return w.WEECHAT_RC_OK

### OPTIONS ###

invertdict = lambda d: dict(zip(d.itervalues(), d.keys()))
booleans = {'on': True, 'off': False}
boolean_options = ['summarize_by_buffer', 'short_buffer_name', 'clear_visited_buffers']

options = {}
for option in settings.keys():
    if option in boolean_options :
        options[option] = booleans[w.config_get_plugin(option)]
    else:
        options[option] = w.config_get_plugin(option)

def my_config_cb(data, option, value):
    global options

    for boolean_option in boolean_options :
        if option.endswith(boolean_option):
            if value in booleans.keys():
                options[boolean_option] = booleans[w.config_get_plugin(boolean_option)]
            else:
                w.prnt('', 'Error: "%s" is not a boolean, please use "on" or "off"' % w.config_get_plugin(boolean_option))
                w.config_set_plugin(boolean_option, invertdict(booleans)[options[boolean_option]])
    write_file()
    return w.WEECHAT_RC_OK

for option in settings.keys():
    w.hook_config("plugins.var.python.%s.%s" % (name, option), "my_config_cb", "")

###  HOOKS  ###

w.hook_command("hl2file_clear", "", "", "", "", "clear_file_cb", "")
w.hook_signal("buffer_switch","buffer_switch_cb","")
w.hook_signal("window_switch","buffer_switch_cb","")
w.hook_print("", "", "", 1, "my_print_cb", "")



"""
Simple weechat script for showing currently playing song from clementine.

Basic install / usage:
place into ~/.weechat/python/autoload
/python load python/autoload/clemenshow.py
/np
"""
SCRIPT_NAME    = "clemenshow"
SCRIPT_AUTHOR  = "Leigh MacDonald <leigh.macdonald@gmail.com>"
SCRIPT_VERSION = "1.0"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC    = "Clementine now playing script"
SCRIPT_COMMAND = "np"

import sys
from os.path import exists

try:
    import weechat
except ImportError:
    print "This script must be run under WeeChat 3.4 or better."
    print "Get WeeChat now at: http://www.weechat.org/"
    sys.exit()
try:
    from dbus import Bus, DBusException
except ImportError:
    print "Please install python-dbus"
    sys.exit()

bus = Bus(Bus.TYPE_SESSION)

def get_type(path):
    p = path.split(".")
    return p[len(p)-1].upper()

#@DebugArgs
def np_command(data, buffer, args):
    try:
        c = bus.get_object('org.mpris.clementine', '/Player')
        f = c.GetMetadata()
        weechat.command(buffer, "/me %s | %s [%s@%dkbps/%dHz]" % (f['artist'].encode('UTF-8'),
                                                                  f['title'].encode('UTF-8'),
                                                                  get_type(f['location'].encode('UTF-8')),
                                                                  int(f['audio-bitrate']),
                                                                  int(f['audio-samplerate'])))
    except DBusException:
        weechat.prnt(buffer, "Doesnt look like clementine is running, if it is make sure dbus is running")
    except Exception, err:
        weechat.prnt(buffer, err)
    finally:
        return weechat.WEECHAT_RC_OK

weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, "on_shutdown", "")
weechat.hook_command(SCRIPT_COMMAND, SCRIPT_DESC, "", "", "", "np_command", "")
weechat.prnt("", "%s | %s" % (SCRIPT_NAME, SCRIPT_AUTHOR))



# -*- coding: utf-8 -*-
#
# Copyright (c) 2013-2017 by nils_2 <weechatter@arcor.de>
#
# add/del channel(s) to/from autojoin option
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# idea by azizLIGHTS
#
# 2017-01-06: nils_2, (freenode.#weechat)
#       0.6 : fix problem with non existing server (reported by Niols)
# 2016-12-19: nils_2, (freenode.#weechat)
#       0.5 : fix problem with empty autojoin (reported by Caelum)
# 2016-06-05: nils_2, (freenode.#weechat)
#       0.4 : make script python3 compatible
# 2015-11-14: nils_2, (freenode.#weechat)
#       0.3 : fix: problem with (undef) option
# 2014-01-19: nils_2, (freenode.#weechat)
#       0.2 : fix: adding keys to already existing keys failed
# 2013-12-22: nils_2, (freenode.#weechat)
#       0.1 : initial release
#
# requires: WeeChat version 0.3.x
#
# Development is currently hosted at
# https://github.com/weechatter/weechat-scripts

try:
    import weechat,re

except Exception:
    print("This script must be run under WeeChat.")
    print("Get WeeChat now at: http://www.weechat.org/")
    quit()

SCRIPT_NAME     = "autojoinem"
SCRIPT_AUTHOR   = "nils_2 <weechatter@arcor.de>"
SCRIPT_VERSION  = "0.6"
SCRIPT_LICENSE  = "GPL"
SCRIPT_DESC     = "add/del channel(s) to/from autojoin option"

OPTIONS         = { 'sorted'        : ('off','channels will be sorted in autojoin-option. if autojoin-option contains channel-keys, this option will be ignored.'),
                  }

def add_autojoin_cmd_cb(data, buffer, args):
    if args == "":                                                                              # no args given. quit
        return weechat.WEECHAT_RC_OK

    argv = args.strip().split(' ')

#    if (len(argv) <= 1):
#        weechat.prnt(buffer,"%s%s: too few arguments." % (weechat.prefix('error'),SCRIPT_NAME))
#        return weechat.WEECHAT_RC_OK

    server = weechat.buffer_get_string(buffer, 'localvar_server')                               # current server
    channel = weechat.buffer_get_string(buffer, 'localvar_channel')                             # current channel
    buf_type = weechat.buffer_get_string(buffer, 'localvar_type')

    # only "add <servername>" given by user
    if (len(argv) == 2):
        weechat.prnt(buffer,"%s%s: invalid number of arguments." % (weechat.prefix('error'),SCRIPT_NAME))
        return weechat.WEECHAT_RC_OK

    # '-key' keyword in command line?
    if '-key' in argv:
        found_key_word = argv.index('-key')
        key_words = argv[int(found_key_word)+1:]
        # don't use "-key" in argv
        argv = argv[:int(found_key_word)]

    # ADD argument
    if (argv[0].lower() == 'add'):
        # add current channel to autojoin. Only option "add" was given..
        if (len(argv) == 1):
            if server == "" or channel == "" or server == channel or buf_type == "" or buf_type != 'channel':
                weechat.prnt(buffer,"%s%s: current buffer is not a channel buffer." % (weechat.prefix('error'),SCRIPT_NAME))
                return weechat.WEECHAT_RC_OK
            list_of_channels, list_of_current_keys = get_autojoin_list(buffer,server)
            # no channels in option!
            if list_of_channels == 1 and list_of_current_keys == 1:
                ptr_config_autojoin = weechat.config_get('irc.server.%s.autojoin' % server)
                rc = weechat.config_option_set(ptr_config_autojoin,channel,1)
                return weechat.WEECHAT_RC_OK
            if channel in list_of_channels:
                weechat.prnt(buffer,"%s%s: channel '%s' already in autojoin for server '%s'" % (weechat.prefix("error"),SCRIPT_NAME,channel,server))
            else:
                # first char of channel '#' ?
                if channel[0] == '#':
                    if '-key' in args and len(key_words) > 1:
                        weechat.prnt(buffer,"%s%s: too many key(s) for given channel(s) " % (weechat.prefix('error'),SCRIPT_NAME))
                        return weechat.WEECHAT_RC_OK
                    elif '-key' in args and len(key_words) == 1:
                        list_of_channels.insert(0,channel)
                        list_of_current_keys = ','.join(key_words)
                        # strip leading ','
                        if list_of_current_keys[0] == ',':
                            list_of_current_keys = list_of_current_keys.lstrip(',')
                    else:
                        list_of_channels.append(channel)

                    if not set_autojoin_list(server,list_of_channels, list_of_current_keys):
                        weechat.prnt(buffer,"%s%s: set new value for option failed..." % (weechat.prefix('error'),SCRIPT_NAME))
        # server and channels given by user
        elif (len(argv) >= 3):
            server = argv[1]
            list_of_channels = argv[2:]
            if '-key' in args and len(list_of_channels) < len(key_words):
                weechat.prnt(buffer,"%s%s: too many key(s) for given channel(s) " % (weechat.prefix('error'),SCRIPT_NAME))
                return weechat.WEECHAT_RC_OK

            list_of_current_channels,list_of_current_keys = get_autojoin_list(buffer,server)
            # autojoin option is empty
            if list_of_current_channels == 1:
                # no channel -> no key!
                list_of_current_keys = ""
                if '-key' in args:
                    list_of_current_keys = ','.join(key_words)
                    # strip leading ','
                    if list_of_current_keys[0] == ',':
                        list_of_current_keys = list_of_current_keys.lstrip(',')
                if not set_autojoin_list(server,list_of_channels, list_of_current_keys):
                    weechat.prnt(buffer,"%s%s: set new value for option failed..." % (weechat.prefix('error'),SCRIPT_NAME))
            else:
                if '-key' in args:
                    j = 0
                    new_keys = []
                    list_of_new_keys = []
                    for i in list_of_channels:
                        if i not in list_of_current_channels and j <= len(key_words):
#                            weechat.prnt(buffer,"channel: %s, channel key is: '%s'" % (i,key_words[j]))
                            list_of_current_channels.insert(j,i)
                            new_keys.insert(j,key_words[j])
                        j += 1
                    missing_channels = list_of_current_channels
                    list_of_new_keys = ','.join(new_keys)
                    if list_of_current_keys:
                        list_of_current_keys = list_of_new_keys + ',' + list_of_current_keys
                    else:
                        list_of_current_keys = list_of_new_keys
                    # strip leading ','
                    if list_of_current_keys[0] == ',':
                        list_of_current_keys = list_of_current_keys.lstrip(',')
                else:
                    # check given channels with channels already set in option
                    missing_channels = get_difference(list_of_channels,list_of_current_channels)
                    missing_channels = list_of_current_channels + missing_channels

                if not set_autojoin_list(server,missing_channels, list_of_current_keys):
                    weechat.prnt(buffer,"%s%s: set new value for option failed..." % (weechat.prefix('error'),SCRIPT_NAME))
        return weechat.WEECHAT_RC_OK

    # DEL argument
    if (argv[0].lower() == 'del'):
        # del current channel from autojoin. Only option "del" was given..
        if (len(argv) == 1):
            if server == "" or channel == "" or server == channel or buf_type == "" or buf_type != 'channel':
                weechat.prnt(buffer,"%s%s: current buffer is not a channel buffer." % (weechat.prefix('error'),SCRIPT_NAME))
                return weechat.WEECHAT_RC_OK
            list_of_channels, list_of_keys = get_autojoin_list(buffer,server)
            # no channels in option, nothing to delete
            if list_of_channels == 1 and list_of_current_keys == 1:
                return weechat.WEECHAT_RC_OK
            if channel not in list_of_channels:
                weechat.prnt(buffer,"%s%s: channel '%s' not found in autojoin for server '%s'" % (weechat.prefix("error"),SCRIPT_NAME,channel,server))
                return weechat.WEECHAT_RC_OK
            else:
                # first char of channel '#' ?
                if channel[0] == '#':
                    channel_key_index = list_of_channels.index(channel)
                    if not list_of_keys:
                        list_of_channels.remove(list_of_channels[channel_key_index])
                        list_of_current_keys = ''
                    else:
                        list_of_keys_tup = list_of_keys.split(",")
                        list_of_current_keys = list_of_keys
                        # channel does not have a key (position of channel > number of keys!)
                        if channel_key_index + 1 > len(list_of_keys_tup):
                            list_of_channels.remove(list_of_channels[channel_key_index])
                        # remove channel and key from autjoin option
                        else:
                            list_of_channels.remove(list_of_channels[channel_key_index])
                            list_of_keys_tup.remove(list_of_keys_tup[channel_key_index])
                            # does a key exists, after removing?
                            if len(list_of_keys_tup) > 0:
                                list_of_current_keys = ','.join(list_of_keys_tup)
                                # strip leading ','
                                if list_of_current_keys[0] == ',':
                                    list_of_current_keys = list_of_current_keys.lstrip(',')
                            else:   # all keys deleted
                                list_of_current_keys = ''

                    # unset option if everything is gone.
                    if not list_of_channels and not list_of_current_keys:
                        ptr_config_autojoin = weechat.config_get('irc.server.%s.autojoin' % server)
                        if ptr_config_autojoin:
                            rc = weechat.config_option_unset(ptr_config_autojoin)
                        return weechat.WEECHAT_RC_OK
                    
                    if not set_autojoin_list(server,list_of_channels, list_of_current_keys):
                        weechat.prnt(buffer,"%s%s: set new value for option failed..." % (weechat.prefix('error'),SCRIPT_NAME))

        # server and channels given by user
        elif (len(argv) >= 3):
            server = argv[1]
            list_of_current_channels,list_of_current_keys = get_autojoin_list(buffer,server)

            # autojoin option is empty
            if list_of_current_channels == 1:
                weechat.prnt(buffer,"%s%s: nothing to delete..." % (weechat.prefix('error'),SCRIPT_NAME))
                return weechat.WEECHAT_RC_OK
            else:
                list_of_channels = args.split(" ")[2:]
                if list_of_current_keys:
                    list_of_current_keys_tup = list_of_current_keys.split(",")
                else:
                    list_of_current_keys_tup = ''

                for i in list_of_channels:
                    # check if given channel is in list of options
                    if not i in list_of_current_channels:
                        continue
                    channel_key_index = list_of_current_channels.index(i)
                    # channel does not have a key (position of channel > number of keys!)
                    if channel_key_index + 1 > len(list_of_current_keys_tup):
                        list_of_current_channels.remove(i)
#                        if len(list_of_current_channels) <= 0:
#                            list_of_current_channels = ''
                    else: # remove channel and key from autjoin option
                        list_of_current_channels.remove(i)
                        list_of_current_keys_tup.remove(list_of_current_keys_tup[channel_key_index])
                        # does an key exists, after removing?
                        if len(list_of_current_keys_tup) > 0:
                            list_of_current_keys = ','.join(list_of_current_keys_tup)
                            # strip leading ','
                            if list_of_current_keys[0] == ',':
                                list_of_current_keys = list_of_current_keys.lstrip(',')
                        else:   # all keys deleted
                            list_of_current_keys = ''

#                for j in list_of_current_channels:
#                    weechat.prnt(buffer,"chan:%s" % j)
#                for j in list_of_current_keys_tup:
#                    weechat.prnt(buffer,"key :%s" % j)

                # unset option if everything is gone.
                if not list_of_current_channels and not list_of_current_keys:
                    ptr_config_autojoin = weechat.config_get('irc.server.%s.autojoin' % server)
                    if ptr_config_autojoin:
                        rc = weechat.config_option_unset(ptr_config_autojoin)
                    return weechat.WEECHAT_RC_OK

                if not set_autojoin_list(server,list_of_current_channels, list_of_current_keys):
                    weechat.prnt(buffer,"%s%s: set new value for option failed..." % (weechat.prefix('error'),SCRIPT_NAME))

    return weechat.WEECHAT_RC_OK

def get_difference(list1, list2):
    return list(set(list1).difference(set(list2)))

# returns a list of channels and a list of keys
# 1 = something failed, 0 = channel found
def get_autojoin_list(buffer,server):
    ptr_config_autojoin = weechat.config_get('irc.server.%s.autojoin' % server)
    # option not found! server does not exist
    if not ptr_config_autojoin:
        weechat.prnt("","%s%s: server '%s' does not exist." % (weechat.prefix('error'),SCRIPT_NAME,server))
        return 1,1

    # get value from autojoin option
    channels = weechat.config_string(ptr_config_autojoin)
    if not channels:
        return 1,1

    # check for keys
    if len(re.findall(r" ", channels)) == 0:
        list_of_channels = channels.split(",")
        list_of_keys = []
    elif len(re.findall(r" ", channels)) == 1:
        list_of_channels2,list_of_keys = channels.split(" ")
        list_of_channels = list_of_channels2.split(",")
    else:
        weechat.prnt("","%s%s: irc.server.%s.autojoin not valid..." % (weechat.prefix('error'),SCRIPT_NAME,server))
        return 1,1

    return list_of_channels, list_of_keys

def set_autojoin_list(server,list_of_channels, list_of_keys):
    ptr_config_autojoin = weechat.config_get('irc.server.%s.autojoin' % server)
    if not ptr_config_autojoin:
        return 0

    if OPTIONS['sorted'].lower() == 'on' and not list_of_keys:
        # no keys, sort the channel-list
        channels = '%s' % ','.join(sorted(list_of_channels))
    else:
        # don't sort channel-list with given key
        channels = '%s' % ','.join(list_of_channels)

    # strip leading ','
    if channels[0] == ',':
        channels = channels.lstrip(',')

    # add keys to list of channels
    if list_of_keys:
        channels = '%s %s' % (channels,list_of_keys)

    rc = weechat.config_option_set(ptr_config_autojoin,channels,1)
    if not rc:
        return 0
    return 1

def autojoinem_completion_cb(data, completion_item, buffer, completion):
#    server = weechat.buffer_get_string(buffer, 'localvar_server')                               # current buffer
    input_line = weechat.buffer_get_string(buffer, 'input')

    # get information out of the input_line
    argv = input_line.strip().split(" ",3)
    if (len(argv) >= 3 and argv[1] == 'del'):
        server = argv[2]

    list_of_channels,list_of_keys = get_autojoin_list(buffer,server)
    if list_of_channels == 1:
        return weechat.WEECHAT_RC_OK

    if (len(argv) >= 4 and argv[1] == 'del'):
        list_of_current_channels = argv[3].split(' ')
        missing_channels = get_difference(list_of_channels,list_of_current_channels)
        if not missing_channels:
            return weechat.WEECHAT_RC_OK
        list_of_channels = missing_channels

    for i, elem in enumerate(list_of_channels):
        weechat.hook_completion_list_add(completion, list_of_channels[i], 0, weechat.WEECHAT_LIST_POS_END)
    return weechat.WEECHAT_RC_OK
# ================================[ weechat options & description ]===============================
def init_options():
    for option,value in OPTIONS.items():
        weechat.config_set_desc_plugin(option, '%s (default: "%s")' % (value[1], value[0]))
        if not weechat.config_is_set_plugin(option):
            weechat.config_set_plugin(option, value[0])
            OPTIONS[option] = value[0]
        else:
            OPTIONS[option] = weechat.config_get_plugin(option)

def toggle_refresh(pointer, name, value):
    global OPTIONS
    option = name[len('plugins.var.python.' + SCRIPT_NAME + '.'):]        # get optionname
    OPTIONS[option] = value                                               # save new value
    return weechat.WEECHAT_RC_OK
# ================================[ main ]===============================
if __name__ == "__main__":
    if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, '', ''):
        version = weechat.info_get("version_number", "") or 0
        weechat.hook_command(SCRIPT_NAME,SCRIPT_DESC,
                             'add <server> [<channel1>[ <channel2>...]] | [-key <channelkey> [<channelkey>...]] ||'
                             'del <server> [<channel1>[ <channel2>...]]',
                             'add <server> <channel>: add channel to irc.server.<servername>.autojoin\n'
                             '     -key <channelkey>: name of channelkey\n'
                             'del <server> <channel>: del channel from irc.server.<servername>.autojoin\n'
                             '\n'
                             'Examples:\n'
                             ' add current channel to corresponding server option:\n'
                             '  /' + SCRIPT_NAME + ' add\n'
                             ' add all channels from all server to corresponding server option:\n'
                             '  /allchan /' + SCRIPT_NAME + ' add\n'
                             ' add channel #weechat to autojoin option on server freenode:\n'
                             '  /' + SCRIPT_NAME + ' add freenode #weechat\n'
                             ' add channel #weechat and #weechat-de to autojoin option on server freenode, with channel key for channel #weechat:\n'
                             '  /' + SCRIPT_NAME + ' add freenode #weechat #weechat-de -key my_channel_key\n'
                             ' del channels #weechat and #weechat-de from autojoin option on server freenode:\n'
                             '  /' + SCRIPT_NAME + ' del freenode #weechat #weechat-de',
                             'add %(irc_servers) %(irc_server_channels)|%*||'
                             'del %(irc_servers) %(plugin_autojoinem)|%*',
                             'add_autojoin_cmd_cb', '')

        init_options()
        weechat.hook_completion('plugin_autojoinem', 'autojoin_completion', 'autojoinem_completion_cb', '')
        weechat.hook_config('plugins.var.python.' + SCRIPT_NAME + '.*', 'toggle_refresh', '')

#        if int(version) >= 0x00030600:
#        else:
#            weechat.prnt("","%s%s %s" % (weechat.prefix("error"),SCRIPT_NAME,": needs version 0.3.6 or higher"))
#            weechat.command("","/wait 1ms /python unload %s" % SCRIPT_NAME)
